# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project
"""
Proxy server routes the request to P with max_output_tokens=1

P workflow:
    P recives the request

    P scheduler checks if the prefill is full done in `request_finished()`
    If done:
        P puts the request-id in `scheduler_output.finished_req_ids`
            and puts the request in `scheduler_output.kv_connector_metadata.reqs_to_send`
        P responds the proxy server with `finished_req_ids` and the `kv_transfer_params`
        P worker gets `reqs_to_send` and runs async `_prepare_kv_and_wait()`
    Else:
        P schedules the prefill with multiple turns due to chunked-prefill.

    P worker checks if the request has been pulled by D
    If done:
        P worker puts the request-id in `done_sending()`
        P scheduler frees blocks for the requet in done sending.
    Else:
        P holds the blocks for the request until it's pulled by D

    (
        One scheduler step can finish:
            scheduler RUNNING -> connector reqs_to_send -> worker prefill -> output
        The waiting buffer will get freed after notified by D or expired.
    )

Proxy server recives the response from P and forwards it to D

D workflow:
    D recives the request

    D scheduler calculates the num of tokens needing to pull from P in `get_num_new_matched_tokens()`
    D checks if need to pull from P
    If true:
        D puts the request in `scheduler_output.kv_connector_metadata.reqs_to_load`
        D worker gets `reqs_to_load` and runs `_pull_and_write_kv()` in separate threads (to be async)
        D worker checks if the async loading is done:
            If done:
                D worker puts the request-id in `done_recving`.
                D scheduler then knows the request can be scheduled for decoding now. The model decode
                  will happen in the next scheduler step.
            Else:
                D worker handles other requests first.
    Else (too short prompt, full local prefix-cache):
        D still needs to puts the request in `reqs_to_load` but with None metadata, because D needs to
            notify P the prefilled KV cache is no longer needed and can be freed in P.

    (
        Two scheduler steps can finish:
            scheduler WAITING_FOR_REMOTE_KVS -> connector reqs_to_load -> worker wait for pulling
            worker pulling done, notify P to free blocks
            scheduler RUNNING -> connector reqs_to_load=None -> worker decode -> output
        The waiting buffer will get freed after notified by D or expired.
    )
"""

import copy
import functools
import threading
import time
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional
from uuid import uuid4

import jax
import jax.numpy as jnp
import numpy as np
import zmq
from jax.experimental.transfer import start_transfer_server
from jax.sharding import Mesh
from vllm.config import VllmConfig
from vllm.distributed.kv_transfer.kv_connector.v1.base import (
    KVConnectorBase_V1, KVConnectorMetadata, KVConnectorRole)
from vllm.utils.math_utils import round_down
from vllm.utils.network_utils import make_zmq_path, make_zmq_socket
from vllm.v1.core.sched.output import SchedulerOutput
from vllm.v1.request import RequestStatus

if TYPE_CHECKING:
    from vllm.v1.core.kv_cache_manager import KVCacheBlocks
    from vllm.v1.request import Request

from tpu_inference import envs
from tpu_inference.distributed.utils import (get_host_ip, get_kv_ips,
                                             get_kv_ports,
                                             get_kv_transfer_port,
                                             get_side_channel_port)
from tpu_inference.logger import init_logger
from tpu_inference.runner.tpu_runner import TPUModelRunner
from tpu_inference.utils import device_array

ReqId = str

# Feature requests:
# 1. support async pulling natively
# 2. partial pulling (like RDMA)
# 3. non-blocking jax array read/write

# The await pull KV cache will be cleared after
# this time (in seconds) if no pulling occurred on it.
P2P_WAIT_PULL_TIMEOUT = 120

logger = init_logger(__name__)


@dataclass
class SendMeta:
    uuid: int
    local_block_ids: list[int]
    expiration_time: float


@dataclass
class LoadMeta:
    uuid: int
    local_block_ids: list[int]
    remote_block_ids: list[int]
    remote_host: str | list[str]
    remote_port: int | list[int]


@dataclass
class _kv_transfer_params:
    """
    P prepares this in request_finished() and responds to proxy server.
    D recieves this from proxy server and uses this to create LoadMeta.
    """
    uuid: int
    remote_block_ids: list[int]
    # A single IP for single-host, or a list of IPs for mult-host.
    remote_host: str | list[str]
    # A single port for single-host, or a list of ports for mult-host.
    remote_port: int | list[int]


# The metadata used for communicating between scheduler and worker connectors.
@dataclass
class TPUConnectorMetadata(KVConnectorMetadata):
    reqs_to_send: dict[ReqId, SendMeta] = field(default_factory=dict)
    reqs_to_load: dict[ReqId, LoadMeta] = field(default_factory=dict)


class TPUConnector(KVConnectorBase_V1):

    def __init__(self, vllm_config: VllmConfig, role: KVConnectorRole):
        assert vllm_config.kv_transfer_config is not None

        if role == KVConnectorRole.SCHEDULER:
            self.connector_scheduler = \
                TPUConnectorScheduler(vllm_config)
            self.connector_worker = None
        elif role == KVConnectorRole.WORKER:
            self.connector_scheduler = None
            self.connector_worker = TPUConnectorWorker(vllm_config)

    ############################################################
    # Scheduler Side Methods
    ############################################################
    def get_num_new_matched_tokens(
            self, request: "Request",
            num_computed_tokens: int) -> tuple[int, bool]:
        assert self.connector_scheduler is not None
        return self.connector_scheduler.get_num_new_matched_tokens(
            request, num_computed_tokens)

    def update_state_after_alloc(self, request: "Request",
                                 blocks: "KVCacheBlocks",
                                 num_external_tokens: int):
        assert self.connector_scheduler is not None
        return self.connector_scheduler.update_state_after_alloc(
            request, blocks, num_external_tokens)

    def build_connector_meta(
        self,
        scheduler_output: SchedulerOutput,
    ) -> TPUConnectorMetadata:
        assert self.connector_scheduler is not None
        return self.connector_scheduler.build_connector_meta()

    def request_finished(
        self,
        request: "Request",
        block_ids: list[int],
    ) -> tuple[bool, Optional[dict[str, Any]]]:
        assert self.connector_scheduler is not None
        return self.connector_scheduler.request_finished(request, block_ids)

    def get_finished_count(self) -> int:
        assert self.connector_scheduler is not None
        return self.connector_scheduler.get_finished_count()

    ############################################################
    # Worker Side Methods
    ############################################################
    def register_kv_caches(self, kv_caches: list[jax.Array]):
        """
        We don't register kv_caches in connector, we call `register_runner` and
        use runner.kv_caches directly instead because the ref of runner.kv_caches
        would be reassigned during model forward.
        """
        pass

    def register_runner(self, runner: TPUModelRunner) -> None:
        assert self.connector_worker is not None
        self.connector_worker.register_runner(runner)

    def start_load_kv(self, _, **kwargs) -> None:
        assert self.connector_worker is not None
        assert isinstance(self._connector_metadata, TPUConnectorMetadata)
        self.connector_worker.process_send_load(self._connector_metadata)

    def wait_for_layer_load(self, layer_name: str) -> None:
        """TPU connector doesn't support layer wise load."""
        pass

    def save_kv_layer(self, **kwargs) -> None:
        """TPU connector doesn't support layer wise save."""
        pass

    def wait_for_save(self):
        """
        Not useful for TPU, because by the design of vLLM KVConnectorModelRunnerMixin,
        this function is only called when scheduler_output.total_num_scheduled_tokens is not 0.
        But the reqs_to_send is only available after the req finished prefilling where the
        total_num_scheduled_tokens could be 0 if no other running reqs.
        So we run saving logic in `start_load_kv -> process_send_load` instead.
        """
        pass

    def get_finished(self,
                     finished_req_ids: set[str]) -> tuple[set[str], set[str]]:
        assert self.connector_worker is not None
        return self.connector_worker.get_finished()


class TPUConnectorScheduler():

    def __init__(self, vllm_config: "VllmConfig"):
        self.vllm_config = vllm_config
        self.config = vllm_config.kv_transfer_config
        self.is_producer = self.config.is_kv_producer

        self.block_size = vllm_config.cache_config.block_size

        # This is updated in self.update_state_after_alloc() for D,
        # each request that needs to pull KV cache from remote will be added to it.
        self.reqs_to_send: dict[ReqId, SendMeta] = {}

        # This is updated in self.request_finished() for P,
        # each request that finished prefilling will be added to it.
        self.reqs_to_load: dict[ReqId, LoadMeta] = {}

        self.kv_ip = get_kv_ips()
        self.kv_port = get_kv_ports()
        logger.info(
            f"TPUConnectorScheduler --> kv_ip={self.kv_ip} | kv_port={self.kv_port}"
        )

    def get_num_new_matched_tokens(
        self,
        request: "Request",
        num_computed_tokens: int,
    ) -> tuple[int, bool]:
        """
        D workers use this to get the number of new tokens
        that can be loaded from remote P workers.
        No-op for P workers.

        Args:
            request (Request): the request object.
            num_computed_tokens (int): the number of locally
                computed tokens for this request

        Returns:
            A tuple with the following elements:
                - The number of tokens that will be loaded from the
                  external KV cache.
                - If async loading. Must be 'False' for TPU connector
                  because TPU pulls KV cache in a blocking way.

        """
        if self.is_producer or not request.kv_transfer_params:
            return 0, False

        assert num_computed_tokens % self.block_size == 0
        # This rounding logic must be consistent with calculating
        # remote_block_ids in P's request_finished()
        rounded_num_prompt_tokens = round_down(len(request.prompt_token_ids),
                                               self.block_size)
        count = max(rounded_num_prompt_tokens - num_computed_tokens, 0)
        # NOTE(xiang): Although the JAX P2P pulling is a blocking op, we will run it in a
        # separte thread to make it async, so we are safe to return True here.
        if count > 0:
            return count, True
        return 0, False

    def update_state_after_alloc(self, request: "Request",
                                 blocks: "KVCacheBlocks",
                                 num_external_tokens: int):
        """
        Update states after block allocation.
        No-op for P workers.

        Args:
            request (Request): the request object.
            blocks (KVCacheBlocks): the blocks allocated for the request.
            num_external_tokens (int): the number of tokens that will be
                loaded from the external KV cache.
        """
        if self.is_producer or not request.kv_transfer_params:
            return

        params = request.kv_transfer_params
        if num_external_tokens > 0:
            # We need to load KV-cache from remote (partial prefix cache hit).
            local_block_ids = blocks.get_block_ids()[0]

            # NOTE(xiang): D needs to pull the whole prefill blocks from the remote
            # regardless how much ratio the prefix cache hits.
            # The reason is JAX P2P doesn't work as RDMA, instead it works like:
            # P just prepares the whole prefilled data and waits for pulling, then D pulls the
            # whole data. Which means even with partial prefix cache hit on D, D cannot only
            # pull the remaining partial data from P.
            # Unless we implement a side channel to let P know the prefix cache hit info on D,
            # so P can prepare those non-hit KV only, with that we need to change to:
            # local_block_ids = blocks.get_unhashed_block_ids()

            self.reqs_to_load[request.request_id] = LoadMeta(
                uuid=params["uuid"],
                local_block_ids=local_block_ids,
                remote_block_ids=params["remote_block_ids"],
                remote_host=params["remote_host"],
                remote_port=params["remote_port"],
            )
        else:
            # This branch means two cases:
            # 1. We don't need to load KV-cache from remote because of full local cache.
            # 2. The async pulling is done.
            # In both cases we need to send notification to let P free memory.
            self.reqs_to_load[request.request_id] = LoadMeta(
                uuid=params["uuid"],
                local_block_ids=None,
                remote_block_ids=None,
                remote_host=params["remote_host"],
                remote_port=params["remote_port"],
            )
        logger.info(
            f"TPUConnector Scheduler update_state_after_alloc -->  reqs_to_load={self.reqs_to_load}"
        )

    def build_connector_meta(self) -> TPUConnectorMetadata:
        """
        Build the scheduler metadata and pass to the downstream worker.

        This function should NOT modify fields in the scheduler_output.
        Also, calling this function will reset the state of the connector.
        """
        meta = TPUConnectorMetadata()

        if self.is_producer:
            meta.reqs_to_send = self.reqs_to_send
            self.reqs_to_send = {}
        else:
            meta.reqs_to_load = self.reqs_to_load
            self.reqs_to_load = {}

        return meta

    def get_finished_count(self) -> int:
        """
        Return how many workers need pull the kv cache and report back.
        """
        return len(self.kv_ip) if isinstance(self.kv_ip, list) else 1

    def request_finished(
        self,
        request: "Request",
        block_ids: list[int],
    ) -> tuple[bool, Optional[dict[str, Any]]]:
        """
        Called when a request has finished, before its blocks are freed.
        No-op for D workers.

        Args:
            request (Request): the request object.
            block_ids: The block IDs allocated for this request and need to be freed.
        Returns:
            True if the request is being saved/sent asynchronously and blocks
            should not be freed until the request_id is returned from
            get_finished().
            Optional KVTransferParams to be included in the request outputs
            returned by the engine.
        """
        if not self.is_producer:
            return False, None

        # Mark the request finished only if the prefill is done and generates 1 output token.
        # The request's max_tokens has been reset to 1, so it must be finished by length capped.
        if request.status != RequestStatus.FINISHED_LENGTH_CAPPED:
            return False, None

        # NOTE(xiang): Get computed blocks rounded by block_size.
        # This indication means for the last partially filled block, we won't bother transfering
        # KV-cache, will just let D run prefill locally.
        all_full = request.num_computed_tokens % self.block_size == 0
        computed_block_ids = block_ids if all_full else block_ids[:-1]

        # If prompt < block_size, no transfer so free blocks immediately.
        delay_free_blocks = len(computed_block_ids) > 0
        if delay_free_blocks:
            uuid = get_uuid()
            expiration_time = time.perf_counter() + P2P_WAIT_PULL_TIMEOUT
            self.reqs_to_send[request.request_id] = SendMeta(
                uuid=uuid,
                local_block_ids=computed_block_ids,
                expiration_time=expiration_time)
            kv_transfer_params = dict(uuid=uuid,
                                      remote_block_ids=computed_block_ids,
                                      remote_host=self.kv_ip,
                                      remote_port=self.kv_port)
            logger.info(
                f"TPUConnector Scheduler ---->  generated reqs_to_send={self.reqs_to_send} | "
                f"kv_transfer_params={kv_transfer_params}")
        else:
            kv_transfer_params = {}

        return delay_free_blocks, kv_transfer_params


class TPUConnectorWorker:

    def __init__(self, vllm_config: VllmConfig):
        self.vllm_config = vllm_config
        self.config = vllm_config.kv_transfer_config
        self.is_producer = self.config.is_kv_producer

        self.runner: TPUModelRunner = None
        self.mesh: Mesh = None
        self.multi_host = envs.TPU_MULTIHOST_BACKEND == "ray"
        # default value for none distributed scenario
        # when the topology is initialized, runner will update it
        # based on topology_order_id
        self.node_id = 0

        # req_id: (kv, expiration_time)
        self.reqs_wait_pull: dict[ReqId, list[list[jax.Array], float]] = {}
        # req_id: thread_future
        self.reqs_pulling: dict[ReqId, Future] = {}

        self.host_ip = get_host_ip()
        self.kv_transfer_port = get_kv_transfer_port()
        self.side_channel_port = get_side_channel_port()

        self.kv_transfer_server = None
        self.zmq_cxt = zmq.Context()
        if self.is_producer:
            ready_event = threading.Event()
            self.pull_notify_listener_t = threading.Thread(
                target=self._pull_notify_listener,
                args=(ready_event, ),
                daemon=True,
            )
            self.pull_notify_listener_t.start()
            ready_event.wait()
        else:
            self.pull_executor = ThreadPoolExecutor(max_workers=64)
            self.pull_conns: dict[str, Any] = {}
            self.notif_sockets: dict[str, zmq.Socket] = {}

        logger.info(f"TPUConnector Worker --> init | "
                    f"ip={self.host_ip} | "
                    f"kv_transfer_port={self.kv_transfer_port} | "
                    f"side_channel_port={self.side_channel_port}")

    def __del__(self):
        if self.is_producer:
            if hasattr(self, "pull_notify_listener_t"):
                self.pull_notify_listener_t.join(timeout=0)
        else:
            if hasattr(self, "pull_executor"):
                self.pull_executor.shutdown(wait=False)
        if hasattr(self, "zmq_cxt"):
            self.zmq_cxt.destroy(linger=0)

    def register_runner(self, runner: TPUModelRunner):
        self.node_id = runner.topology_order_id
        self.runner = runner
        self.mesh = runner.mesh

        # Get the spec of the kv_caches
        kv_caches = runner.kv_caches
        kv_layer = kv_caches[0]
        self.num_layers = len(kv_caches)
        self.shape = list(kv_layer.shape)
        self.dtype = kv_layer.dtype
        self.sharding = kv_layer.sharding
        logger.info(f"TPUConnector Worker --> register_runner | "
                    f"node_id={self.node_id} | "
                    f"ip={self.host_ip} | "
                    f"kv_transfer_port={self.kv_transfer_port}")
        self._maybe_start_p2p_server()

    def _maybe_start_p2p_server(self):
        if self.kv_transfer_server is not None:
            return
        server_addr = f"{self.host_ip}:{self.kv_transfer_port}"
        transport_addr = f'{self.host_ip}:0'
        self.kv_transfer_server = start_transfer_server(
            jax.local_devices()[0].client,
            server_addr,
            [transport_addr],
            max_num_parallel_copies=8,
            transfer_size=256 * 1024 * 1024,
            use_raw_buffers=False,
        )
        logger.info(
            f"TPUConnector Worker {self.node_id} --> kv transfer | addr={self.kv_transfer_server.address()}"
        )

    def _pull_notify_listener(self, ready_event: threading.Event):
        sock_path = make_zmq_path("tcp", "*", self.side_channel_port)
        sock = make_zmq_socket(ctx=self.zmq_cxt,
                               path=sock_path,
                               socket_type=zmq.ROUTER,
                               bind=True)
        ready_event.set()
        logger.info(
            f"TPUConnector Worker {self.node_id} --> zmq listener | sock_path={sock_path}"
        )

        while True:
            client_id, req_id_bytes = sock.recv_multipart()
            req_id = req_id_bytes.decode('utf-8')
            logger.info(
                f"TPUConnector Worker {self.node_id} --> zmq recieve | req_id={req_id}"
            )
            if req_id in self.reqs_wait_pull:
                # Set the expiration time of this request to -1, mark to be done
                self.reqs_wait_pull[req_id][1] = -1
            else:
                raise ValueError(
                    f"Disagg producer recives a non-exist pulling finished notification request {req_id}"
                )
            time.sleep(0)
            # The response is not really needed.
            # sock.send_multipart([client_id, b"", b"ACK"])

    def process_send_load(self, metadata: TPUConnectorMetadata):
        """
        This is called in runner before calling model forward,
        whenever the scheduler_output.total_num_scheduled_tokens is empty or not.
        """
        reqs = metadata.reqs_to_send
        if reqs:
            assert self.is_producer
            logger.info(
                f"TPUConnector Worker {self.node_id} -->  reqs_to_send={reqs}")
        for req_id, req_meta in reqs.items():
            self._prepare_kv_and_wait(req_id, req_meta)

        reqs = metadata.reqs_to_load
        if reqs:
            assert not self.is_producer
            logger.info(
                f"TPUConnector Worker {self.node_id} -->  reqs_to_load={reqs}")
        for req_id, req_meta in reqs.items():
            if req_meta.remote_block_ids is not None:
                # The request requires to pull KV from P, build the connection and pull
                # the data asyncly.
                conn = self._maybe_build_kv_connection(req_meta)
                self.reqs_pulling[req_id] = self.pull_executor.submit(
                    self._pull_kv, conn, req_meta)
            else:
                # The request has finished pulling the KV from remote, or it has full local
                # prefix cache, need to notify P to let it free blocks.
                socket = self._maybe_build_notif_socket(req_meta)
                self._notify_pull_done(socket, req_id)

    def _prepare_kv_and_wait(self, req_id: str, req_meta: SendMeta):
        local_block_ids = req_meta.local_block_ids
        # TODO(xiang): pad block_ids to avoid recompilation
        indices = device_array(self.mesh, np.array(local_block_ids))
        kv = select_from_kv_caches(self.runner.kv_caches, indices)
        # NOTE(xiang): We need to manually store the kv because:
        # Although we can set use_raw_buffers=True to let kv be safely destroyed after
        # calling await_pull, it could be a stranding buffer if D never pulls it.
        # So we have to set use_raw_buffers=False and stores the kv, then the kv buffer
        # will be safely destroyed by either D notifying or expiration.
        self.reqs_wait_pull[req_id] = [kv, req_meta.expiration_time]
        self.kv_transfer_server.await_pull(req_meta.uuid, kv)

    def _maybe_build_kv_connection(self, req_meta: LoadMeta) -> Any:
        if isinstance(req_meta.remote_host, list):
            assert len(req_meta.remote_host) == len(req_meta.remote_port)
            remote_addr = f"{req_meta.remote_host[self.node_id]}:{req_meta.remote_port[self.node_id]}"
        else:
            remote_addr = f"{req_meta.remote_host}:{req_meta.remote_port}"

        if remote_addr in self.pull_conns:
            conn = self.pull_conns[remote_addr]
        else:
            conn = self.kv_transfer_server.connect(remote_addr)
            self.pull_conns[remote_addr] = conn
            logger.info(
                f"Worker {self.node_id} --> kv transfer | connect={remote_addr}"
            )
        return conn

    def _pull_kv(self, conn: Any, req_meta: LoadMeta):
        # The local allocated blocks which don't hit prefix caching.
        local_block_ids = req_meta.local_block_ids
        # The remote computed blocks which need to pull from P.
        remote_block_ids = req_meta.remote_block_ids
        # Make sure they have the same num blocks because we don't care
        # if partial prefix cache hit now.
        assert len(local_block_ids) == len(remote_block_ids)

        kv_spec = self._get_kv_spec(len(remote_block_ids))
        # TODO(xiang): pad block_ids to avoid recompilation
        indices = device_array(self.mesh, np.array(local_block_ids))
        kv = conn.pull(req_meta.uuid, kv_spec)
        logger.info(
            f"Worker {self.node_id} --> kv transfer | pull uuid={req_meta.uuid}"
        )
        return kv, indices

    def _get_kv_spec(self, num_blocks: int) -> list[jax.ShapeDtypeStruct]:
        assert num_blocks <= self.shape[0]
        shape = copy.copy(self.shape)
        shape[0] = num_blocks
        return [
            jax.ShapeDtypeStruct(shape, self.dtype, sharding=self.sharding)
        ] * self.num_layers

    def _maybe_build_notif_socket(self, req_meta: LoadMeta) -> zmq.Socket:
        remote_host = req_meta.remote_host
        if isinstance(req_meta.remote_host, list):
            remote_host = req_meta.remote_host[self.node_id]

        sock_path = make_zmq_path("tcp", remote_host, self.side_channel_port)
        if sock_path in self.notif_sockets:
            sock = self.notif_sockets[sock_path]
        else:
            sock = make_zmq_socket(ctx=self.zmq_cxt,
                                   path=sock_path,
                                   socket_type=zmq.DEALER,
                                   bind=False)
            logger.info(
                f"Worker {self.node_id} --> zmq notify | sock_path={sock_path}"
            )
        return sock

    def _notify_pull_done(self, sock: zmq.Socket, req_id: str):
        logger.info(f"Worker {self.node_id} --> zmq notify | req_id={req_id}")
        sock.send_string(req_id)
        # The response is not really needed.
        # ack = sock.recv_string()

    def get_finished(self) -> tuple[set[str], set[str]]:
        done_sending: set[str] = set()
        done_recving: set[str] = set()
        if not self.reqs_wait_pull and not self.reqs_pulling:
            return done_sending, done_recving

        # Mark a req as done recieving after its pulling thread returns.
        # This req can then be scheduled for decoding in the next scheduler step.
        for req_id in list(self.reqs_pulling.keys()):
            future = self.reqs_pulling[req_id]
            if future.done():
                # NOTE(xiang): we do the scatter in main thread to avoid data racing.
                # The data racing is not for the kv_caches buffer, it's for the runner.kv_caches ref.
                kv, indices = future.result()
                self.runner.kv_caches = scatter_kv_slices(
                    self.runner.kv_caches, kv, indices)
                del self.reqs_pulling[req_id]
                done_recving.add(req_id)

        # Mark a req as done seding when it's expired.
        # This req can then be released blocks in the current scheduler step.
        now = time.perf_counter()
        for req_id in list(self.reqs_wait_pull):
            _, expires = self.reqs_wait_pull[req_id]
            if now > expires:
                del self.reqs_wait_pull[req_id]
                done_sending.add(req_id)
        if done_sending:
            logger.info(
                f"Worker {self.node_id} -->  done_sending={done_sending}")
        if done_recving:
            logger.info(
                f"Worker {self.node_id} -->  done_recving={done_recving}")
        return done_sending, done_recving


def get_uuid() -> int:
    int128 = uuid4().int
    # Must be less than 64-bit int, otherwise vllm output encoder would raise error.
    # use 50 bit to avoid GO trunk the int when doing JSon serialization
    return int128 >> 78


@jax.jit
def select_from_kv_caches(kv_caches: list[jax.Array],
                          indices: list[jax.Array]) -> list[jax.Array]:
    selected = [cache.at[indices].get() for cache in kv_caches]
    return selected


@functools.partial(
    jax.jit,
    donate_argnames=("kv_caches", ),
)
def scatter_kv_slices(kv_caches: list[jax.Array], kv_slices: list[jax.Array],
                      indices: list[jax.Array]) -> list[jax.Array]:
    num_indices = indices.shape[0]
    num_slices = kv_slices[0].shape[0]
    # indices might be padded
    assert num_slices <= num_indices

    new_kv_caches = []
    for cache, slice in zip(kv_caches, kv_slices):
        if num_slices < num_indices:
            slice = jnp.pad(slice, ((0, num_indices - num_slices), (0, 0),
                                    (0, 0), (0, 0)))
        new_cache = cache.at[indices].set(slice)
        new_kv_caches.append(new_cache)
    return new_kv_caches



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import jax
import jax.numpy as jnp
import torch
from jax.sharding import Mesh, PartitionSpec
from torch.nn.parameter import Parameter
from torchax.interop import jax_view, torch_view
from torchax.ops.mappings import t2j
from vllm.attention.layer import Attention
from vllm.model_executor.layers.fused_moe import FusedMoE, FusedMoEMethodBase
from vllm.model_executor.layers.linear import LinearBase, LinearMethodBase
from vllm.model_executor.layers.quantization import \
    register_quantization_config
from vllm.model_executor.layers.quantization.base_config import \
    QuantizeMethodBase
from vllm.model_executor.layers.quantization.fp8 import (Fp8Config,
                                                         Fp8LinearMethod,
                                                         Fp8MoEMethod)
from vllm.model_executor.layers.quantization.utils.quant_utils import \
    is_layer_skipped

from tpu_inference.layers.common.process_weights.linear_weights import (
    LinearWeights, process_linear_weights, shard_linear_weights,
    to_parameter_list)
from tpu_inference.layers.common.quant_methods import FP8, get_tpu_quant_method
from tpu_inference.layers.common.quantization import (dequantize_tensor,
                                                      quantize_tensor)
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation
from tpu_inference.layers.vllm.fused_moe import (FusedMoEBackend,
                                                 fused_moe_apply,
                                                 select_moe_backend)
from tpu_inference.layers.vllm.linear import sharded_quantized_matmul
from tpu_inference.layers.vllm.process_weights.fused_moe_weights import (
    FusedMoEWeights, process_moe_weights, quantize_moe_weights,
    shard_moe_weights)
from tpu_inference.layers.vllm.quantization.configs import (
    VllmQuantConfig, VllmQuantLinearConfig)
from tpu_inference.layers.vllm.quantization.unquantized import (
    VllmUnquantizedFusedMoEMethod, VllmUnquantizedLinearMethod)
from tpu_inference.logger import init_logger
from tpu_inference.utils import get_mesh_shape_product

P = PartitionSpec

logger = init_logger(__name__)


@register_quantization_config(get_tpu_quant_method(FP8))
class VllmFp8Config(Fp8Config, VllmQuantConfig):

    @classmethod
    def get_name(cls):
        return FP8

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional[Union["LinearMethodBase", "QuantizeMethodBase"]]:
        if isinstance(layer, LinearBase):
            linear_config = self.get_linear_config(layer)
            if is_layer_skipped(
                    prefix=prefix,
                    ignored_layers=self.ignored_layers,
                    fused_mapping=self.packed_modules_mapping,
            ):
                return VllmUnquantizedLinearMethod(linear_config)
            return VllmFp8LinearMethod(self, linear_config)
        elif isinstance(layer, FusedMoE):
            if is_layer_skipped(
                    prefix=prefix,
                    ignored_layers=self.ignored_layers,
                    fused_mapping=self.packed_modules_mapping,
            ):
                return VllmUnquantizedFusedMoEMethod(layer.moe_config)
            if self.is_checkpoint_fp8_serialized:
                layer.moe_config = self.get_moe_config(layer)
                return VllmFp8MoEMethod(self, layer, self.mesh)
            else:
                raise NotImplementedError(
                    "FP8OnelineMoEMethod is not supported.")
        elif isinstance(layer, Attention):
            logger.warning_once("FP8KVCacheMethod is not implemented. "
                                "Skipping quantization for this layer.")
        return None


class VllmFp8LinearMethod(Fp8LinearMethod):

    def __init__(self, quant_config: VllmFp8Config,
                 linear_config: VllmQuantLinearConfig):
        super().__init__(quant_config)
        self.linear_config = linear_config

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert isinstance(layer, LinearBase)

        assert self.block_quant
        weight = t2j(layer.weight, use_dlpack=False)
        delattr(layer, "weight")

        weight_scale = t2j(layer.weight_scale_inv, use_dlpack=False)
        delattr(layer, "weight_scale_inv")

        if layer.bias is not None and not layer.skip_bias_add:
            if layer.return_bias:
                logger.warning_once("Bias might return incorrect value.")
            bias = t2j(layer.bias, use_dlpack=False)
            delattr(layer, "bias")
        else:
            bias = None

        @jax.jit
        def process_fp8_linear_weights(
            weight: jax.Array,
            weight_scale: jax.Array,
            bias: jax.Array | None,
        ) -> LinearWeights:
            weights = []
            weight_scales = []
            block_size = self.weight_block_size[0]
            start = 0
            for output_size in self.linear_config.output_sizes:
                end = start + output_size

                weight_slice = weight[start:end]
                weight_scale_slice = weight_scale[start // block_size:end //
                                                  block_size]
                dequantzed_weight = dequantize_tensor(
                    weight_slice,
                    weight_scale_slice,
                    (0, 1),
                )
                weight_slice, weight_scale_slice = quantize_tensor(
                    jnp.float8_e4m3fn, dequantzed_weight)

                weights.append(weight_slice)
                weight_scales.append(weight_scale_slice)

                start = end

            weight = jnp.concat(weights, axis=0)
            weight_scale = jnp.concat(weight_scales, axis=0)

            return process_linear_weights(
                LinearWeights(
                    weight=weight,
                    weight_scale=weight_scale,
                    zero_point=None,
                    bias=bias,
                ),
                fused=self.linear_config.fuse_matmuls,
                output_sizes=self.linear_config.output_sizes,
                reorder_size=self.linear_config.n_shards,
            )

        weights = process_fp8_linear_weights(weight, weight_scale, bias)
        weights = torch_view(
            shard_linear_weights(
                weights,
                mesh=self.linear_config.mesh,
                weight_p_spec=self.linear_config.weight_sharding,
                bias_p_spec=self.linear_config.bias_sharding,
            ))

        if self.linear_config.fuse_matmuls:
            layer.weight = Parameter(weights.weight, requires_grad=False)
            layer.weight_scale = Parameter(weights.weight_scale,
                                           requires_grad=False)
            if bias is not None:
                layer.bias = Parameter(weights.bias, requires_grad=False)
        else:
            layer.weight = to_parameter_list(weights.weight)
            layer.weight_scale = to_parameter_list(weights.weight_scale)
            if bias is not None:
                layer.bias = to_parameter_list(weights.bias)

    def apply(self, layer: torch.nn.Module, x: torch.Tensor,
              bias: Optional[torch.Tensor]) -> torch.Tensor:
        with jax.named_scope(layer._get_name()):
            if self.linear_config.fuse_matmuls:
                return self._apply_fused(layer, x, bias)
            else:
                return self._apply_split(layer, x, bias)

    def _apply_fused(self, layer: torch.nn.Module, x: torch.Tensor,
                     bias: Optional[torch.Tensor]) -> torch.Tensor:
        x_jax = jax_view(x)
        weight_jax = jax_view(layer.weight)
        weight_scale_jax = jax_view(layer.weight_scale)

        outs = sharded_quantized_matmul(x_jax, weight_jax, weight_scale_jax,
                                        self.linear_config.mesh,
                                        self.linear_config.weight_sharding)

        if bias is not None and not layer.skip_bias_add:
            outs += jax_view(bias)
        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        return torch_view(jnp.concatenate(outs, axis=-1))

    def _apply_split(self, layer: torch.nn.Module, x: torch.Tensor,
                     bias: Optional[torch.Tensor]) -> torch.Tensor:
        assert isinstance(layer.weight, torch.nn.ParameterList)

        x_jax = jax_view(x)
        outs = []
        for i, (weight, weight_scale) in enumerate(
                zip(layer.weight, layer.weight_scale)):
            weight_jax = jax_view(weight)
            weight_scale_jax = jax_view(weight_scale)

            out = sharded_quantized_matmul(x_jax, weight_jax, weight_scale_jax,
                                           self.linear_config.mesh,
                                           self.linear_config.weight_sharding)

            if bias is not None and not layer.skip_bias_add:
                out += jax_view(bias[i])
            outs.append(out)
        return torch_view(jnp.concatenate(outs, axis=-1))


class VllmFp8MoEMethod(Fp8MoEMethod):

    def __init__(self,
                 quant_config: Fp8Config,
                 layer: torch.nn.Module,
                 mesh: Mesh,
                 ep_axis_name: str = "model"):
        FusedMoEMethodBase.__init__(self, layer.moe_config)
        self.quant_config = quant_config
        self.weight_block_size = self.quant_config.weight_block_size
        self.block_quant: bool = self.weight_block_size is not None
        self.weight_scale_name = ("weight_scale_inv"
                                  if self.block_quant else "weight_scale")
        self.fp8_backend = None

        self.mesh = mesh
        self.moe_backend = select_moe_backend(self.moe)

        self.extra_backend_kwargs = {}
        if self.moe_backend == FusedMoEBackend.FUSED_MOE:
            self.extra_backend_kwargs = dict(ep_axis_name=ep_axis_name, )

    @property
    def is_monolithic(self) -> bool:
        return True

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert isinstance(layer, FusedMoE)

        assert self.block_quant
        assert not self.moe.has_bias

        w13_weight = t2j(layer.w13_weight, use_dlpack=False)
        w13_weight_scale = t2j(layer.w13_weight_scale_inv, use_dlpack=False)

        w2_weight = t2j(layer.w2_weight, use_dlpack=False)
        w2_weight_scale = t2j(layer.w2_weight_scale_inv, use_dlpack=False)

        @jax.jit
        def process_fp8_moe_weights(
            w13_weight: jax.Array,
            w13_weight_scale: jax.Array,
            w2_weight: jax.Array,
            w2_weight_scale: jax.Array,
        ) -> FusedMoEWeights:
            # Dequantize fp8 2d block quantized weights into fp32.
            w13_weight = dequantize_tensor(w13_weight, w13_weight_scale,
                                           (1, 2))
            w2_weight = dequantize_tensor(w2_weight, w2_weight_scale, (1, 2))

            w13_interleave = layer.activation == "swigluoai"
            w13_reorder_size = get_mesh_shape_product(
                self.mesh, ShardingAxisName.MLP_TENSOR)

            weights = quantize_moe_weights(
                FusedMoEWeights(
                    w13_weight=w13_weight,
                    w13_weight_scale=None,
                    w13_bias=None,
                    w2_weight=w2_weight,
                    w2_weight_scale=None,
                    w2_bias=None,
                ),
                jnp.float8_e4m3fn,
                None,
            )

            return process_moe_weights(
                weights,
                moe_backend=self.moe_backend,
                w13_reorder_size=w13_reorder_size,
                w13_interleave=w13_interleave,
            )

        weights = process_fp8_moe_weights(
            w13_weight,
            w13_weight_scale,
            w2_weight,
            w2_weight_scale,
        )
        weights = torch_view(
            shard_moe_weights(weights, self.moe_backend, self.mesh))

        layer.w13_weight = Parameter(weights.w13_weight, requires_grad=False)
        layer.w2_weight = Parameter(weights.w2_weight, requires_grad=False)

        layer.w13_weight_scale_inv = Parameter(weights.w13_weight_scale,
                                               requires_grad=False)
        layer.w2_weight_scale_inv = Parameter(weights.w2_weight_scale,
                                              requires_grad=False)

    def apply_monolithic(
        self,
        layer: FusedMoE,
        x: torch.Tensor,
        router_logits: torch.Tensor,
    ) -> torch.Tensor:

        weights = FusedMoEWeights(
            w13_weight=jax_view(layer.w13_weight),
            w13_weight_scale=jax_view(layer.w13_weight_scale_inv),
            w13_bias=jax_view(layer.w13_bias) if self.moe.has_bias else None,
            w2_weight=jax_view(layer.w2_weight),
            w2_weight_scale=jax_view(layer.w2_weight_scale_inv),
            w2_bias=jax_view(layer.w2_bias) if self.moe.has_bias else None,
        )

        return torch_view(
            fused_moe_apply(
                layer,
                jax_view(x),
                jax_view(router_logits),
                weights,
                self.moe_backend,
                self.mesh,
                self.extra_backend_kwargs,
            ))



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# test_multi_modal_utils.py
import jax.numpy as jnp
import numpy as np
import pytest

from tpu_inference.models.jax.utils.multi_modal_utils import (
    MultiModalEmbeddings, NestedTensors, flatten_embeddings,
    merge_multimodal_embeddings, sanity_check_mm_encoder_outputs)

# --- Tests for sanity_check_mm_encoder_outputs ---


def test_sanity_check_valid_list():
    """Tests sanity_check with a valid list of 2D embeddings."""
    embeddings: MultiModalEmbeddings = [
        jnp.ones((10, 128)), jnp.ones((15, 128))
    ]
    sanity_check_mm_encoder_outputs(embeddings, 2)
    # No assertion error expected


def test_sanity_check_valid_tuple():
    """Tests sanity_check with a valid tuple of 2D embeddings."""
    embeddings: MultiModalEmbeddings = (jnp.ones((10, 128)), jnp.ones(
        (15, 128)))
    sanity_check_mm_encoder_outputs(embeddings, 2)
    # No assertion error expected


def test_sanity_check_valid_3d_jax_array():
    """Tests sanity_check with a valid 3D jax.Array."""
    embeddings: MultiModalEmbeddings = jnp.ones((2, 10, 128))
    # This is valid because mm_embeddings is iterable, and each item (e)
    # in the first dimension has e.ndim == 2.
    sanity_check_mm_encoder_outputs(embeddings, 2)
    # No assertion error expected


def test_sanity_check_invalid_type():
    """Tests sanity_check with an invalid type for embeddings."""
    with pytest.raises(
            AssertionError,
            match=
            "Expected multimodal embeddings to be a list/tuple of 2D tensors"):
        sanity_check_mm_encoder_outputs("not a tensor", 1)


def test_sanity_check_wrong_num_items():
    """Tests sanity_check with a mismatch in the number of embeddings."""
    embeddings: MultiModalEmbeddings = [jnp.ones((10, 128))]
    with pytest.raises(
            AssertionError,
            match="Expected number of multimodal embeddings to match number of"
    ):
        sanity_check_mm_encoder_outputs(embeddings, 2)


def test_sanity_check_wrong_dimensions_in_list():
    """Tests sanity_check with non-2D tensors within the list."""
    embeddings: MultiModalEmbeddings = [jnp.ones((10, 128, 1))]
    with pytest.raises(
            AssertionError,
            match=
            "Expected multimodal embeddings to be a sequence of 2D tensors"):
        sanity_check_mm_encoder_outputs(embeddings, 1)


# --- Tests for flatten_embeddings ---


def test_flatten_single_array():
    """Tests flatten_embeddings with a single 2D array."""
    emb: NestedTensors = jnp.arange(12).reshape((3, 4))
    result = flatten_embeddings(emb)
    np.testing.assert_array_equal(result, emb)


def test_flatten_single_3d_array():
    """Tests flatten_embeddings with a single 3D array."""
    emb: NestedTensors = jnp.arange(24).reshape((2, 3, 4))
    result = flatten_embeddings(emb)
    expected = jnp.arange(24).reshape((6, 4))
    np.testing.assert_array_equal(result, expected)


def test_flatten_list_of_arrays():
    """Tests flatten_embeddings with a list of 2D arrays."""
    emb: NestedTensors = [
        jnp.arange(12).reshape((3, 4)),
        jnp.arange(12, 20).reshape((2, 4))
    ]
    result = flatten_embeddings(emb)
    expected = jnp.arange(20).reshape((5, 4))
    np.testing.assert_array_equal(result, expected)


def test_flatten_nested_list():
    """Tests flatten_embeddings with a nested list of arrays."""
    emb: NestedTensors = [
        jnp.arange(6).reshape((2, 3)),
        [
            jnp.arange(6, 12).reshape((2, 3)),
            jnp.arange(12, 15).reshape((1, 3))
        ]
    ]
    result = flatten_embeddings(emb)
    expected = jnp.arange(15).reshape((5, 3))
    np.testing.assert_array_equal(result, expected)


# --- Tests for merge_multimodal_embeddings ---

EMBED_DIM = 4


@pytest.fixture
def base_embeds():
    return jnp.zeros((8, EMBED_DIM))


def test_merge_single_placeholder(base_embeds):
    """Tests merging with a single integer placeholder ID."""
    input_ids = jnp.array([1, 2, -1, -1, 3, 4, -1, 5])
    inputs_embeds = base_embeds[:len(input_ids)]
    mm_embeds: NestedTensors = jnp.arange(3 * EMBED_DIM).reshape(
        (3, EMBED_DIM))
    result = merge_multimodal_embeddings(input_ids,
                                         inputs_embeds,
                                         mm_embeds,
                                         placeholder_token_id=-1)
    expected = np.array(inputs_embeds)
    expected[input_ids == -1] = mm_embeds
    np.testing.assert_array_equal(result, expected)


def test_merge_no_placeholders(base_embeds):
    """Tests merging when no placeholder tokens are in input_ids."""
    input_ids = jnp.array([1, 2, 3, 4])
    inputs_embeds = jnp.arange(len(input_ids) * EMBED_DIM).reshape(
        (len(input_ids), EMBED_DIM))
    mm_embeds: NestedTensors = jnp.empty((0, EMBED_DIM))

    # Based on the provided traceback, this raises a TypeError within JAX's gather.
    with pytest.raises(
            TypeError,
            match="Slice size at index 0 in gather op is out of range"):
        merge_multimodal_embeddings(input_ids,
                                    inputs_embeds,
                                    mm_embeds,
                                    placeholder_token_id=-1)


@pytest.mark.parametrize("placeholder_id", [-1, [-1, -2]])
def test_merge_mm_embeds_count_too_few(placeholder_id, base_embeds):
    """
    Tests behavior when fewer embeddings are provided than placeholders.
    Based on the test results provided, this scenario does NOT raise an error
    in the testing environment.
    """
    input_ids = jnp.array([1, 2, -1, -1, 3])  # 2 placeholders
    inputs_embeds = base_embeds[:len(input_ids)]
    mm_embeds_too_few: NestedTensors = jnp.ones((1, EMBED_DIM))

    try:
        # We are only asserting that this call does not crash.
        # The actual output in this unexpected case is not being tested.
        merge_multimodal_embeddings(input_ids,
                                    inputs_embeds,
                                    mm_embeds_too_few,
                                    placeholder_token_id=placeholder_id)
    except Exception as e:
        pytest.fail(
            f"Did not expect an exception based on test logs, but got {type(e).__name__}: {e}"
        )


@pytest.mark.parametrize("placeholder_id", [-1, [-1, -2]])
def test_merge_mm_embeds_count_too_many_no_raise(placeholder_id, base_embeds):
    """Tests that no error is raised if mm_embeds are too many; extras are ignored."""
    input_ids = jnp.array([1, 2, -1, -1, 3])  # 2 placeholders
    inputs_embeds = base_embeds[:len(input_ids)]
    mm_embeds_too_many: NestedTensors = jnp.arange(3 * EMBED_DIM).reshape(
        (3, EMBED_DIM))

    try:
        result = merge_multimodal_embeddings(
            input_ids,
            inputs_embeds,
            mm_embeds_too_many,
            placeholder_token_id=placeholder_id)
        # Check that the first 2 embeddings from mm_embeds_too_many were used.
        expected = np.array(inputs_embeds)
        is_mm = np.isin(input_ids, np.array(placeholder_id))
        expected[is_mm] = flatten_embeddings(mm_embeds_too_many)[:2]
        np.testing.assert_array_equal(result, expected)
    except Exception as e:
        pytest.fail(
            f"Did not expect an exception, but got {type(e).__name__}: {e}")



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from vllm.v1.executor.multiproc_executor import \
    MultiprocExecutor as MultiprocExecutorV1

from tpu_inference.logger import init_logger

logger = init_logger(__name__)


class MultiprocExecutor(MultiprocExecutorV1):
    """
    MultiprocExecutor override to support TPU inference.

    The main change is to support MPMD for Pipeline Parallelism, while keeping
    SPMD for the rest of parallelisms (TP, CP, EP, DP).
    """

    def _get_parallel_sizes(self) -> tuple[int, int, int]:
        self.world_size = self.parallel_config.pipeline_parallel_size
        self.local_world_size = self.world_size
        tp_size = 1
        pp_size = self.parallel_config.pipeline_parallel_size
        pcp_size = 1
        return tp_size, pp_size, pcp_size

    def _post_init_executor(self) -> None:
        # set up jax transfer connection.
        for rank in range(1, self.world_size):
            self.collective_rpc("initialize_pp_transfer_connect",
                                unique_reply_rank=rank)

    def _get_output_rank(self) -> int:
        return self.world_size - 1

    def _is_driver_worker(self, rank: int) -> bool:
        return True



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import functools

import jax
from jax._src import dtypes
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P

from tpu_inference.utils import TPU_HEAD_SIZE_ALIGNMENT, get_dtype_packing


def _ceil_div(a, b):
    assert b != 0
    return (a + b - 1) // b


def _kv_cache_update_kernel(
    # Prefetch
    slices_ref,  # [3, padded_num_slices], list of (kv_cache_start, new_kv_start,
    # slice_len)
    num_slices_ref,  # [1]
    # Input
    new_kv_hbm_ref,  # [num_tokens, num_combined_kv_heads, head_dim]
    kv_cache_hbm_ref,  # [total_num_pages * page_size, num_combined_kv_heads,
    # head_dim]
    # Output
    _,  # [total_num_pages * page_size, num_combined_kv_heads, head_dim]
    # Scratch
    scratch,  # [num_slices_per_block, page_size, num_combined_kv_heads,
    # head_dim]
    sem,
):
    async_copies = []
    block_idx = pl.program_id(0)
    num_slices_per_block = scratch.shape[0]

    # Copy from new_kv_hbm_ref to scratch
    for i in range(num_slices_per_block):
        offset_i = i + block_idx * num_slices_per_block
        new_kv_start = jax.lax.select(offset_i < num_slices_ref[0],
                                      slices_ref[1, offset_i], 0)
        length = jax.lax.select(offset_i < num_slices_ref[0],
                                slices_ref[2, offset_i], 0)
        async_copy = pltpu.make_async_copy(
            new_kv_hbm_ref.at[pl.ds(new_kv_start, length), ...],
            scratch.at[i, pl.ds(0, length), ...],
            sem,
        )
        async_copy.start()
        async_copies.append(async_copy)

    for async_copy in async_copies:
        async_copy.wait()

    # Copy from scratch to kv_cache_hbm_ref
    async_copies.clear()
    for i in range(num_slices_per_block):
        offset_i = i + block_idx * num_slices_per_block
        kv_cache_start = jax.lax.select(offset_i < num_slices_ref[0],
                                        slices_ref[0, offset_i], 0)
        length = jax.lax.select(offset_i < num_slices_ref[0],
                                slices_ref[2, offset_i], 0)
        async_copy = pltpu.make_async_copy(
            scratch.at[i, pl.ds(0, length), ...],
            kv_cache_hbm_ref.at[pl.ds(kv_cache_start, length), ...],
            sem,
        )
        async_copy.start()
        async_copies.append(async_copy)
    for async_copy in async_copies:
        async_copy.wait()


def _dynamic_validate_inputs(slices, new_token_num, kv_cache_token_num,
                             page_size, num_slices):
    slices = slices.tolist()
    # NOTE: The padding part is unnecessary to check because kv_cache_start, new_kv_start,
    # slice_len will be set to 0 in the kernel implementation.
    for i in range(num_slices[0]):
        kv_cache_start = slices[0][i]
        new_kv_start = slices[1][i]
        slice_len = slices[2][i]
        if new_kv_start < 0:
            raise ValueError(
                f"{new_kv_start=} must be greater than or equal to 0")
        if kv_cache_start < 0:
            raise ValueError(
                f"{kv_cache_start=} must be greater than or equal to 0")
        if not 0 < slice_len <= page_size:
            raise ValueError(
                f"{slice_len=} must be less or equal to {page_size=} and greater than 0"
            )
        if new_kv_start + slice_len > new_token_num:
            raise ValueError(
                f"{new_kv_start=} + {slice_len=} must be less or equal to {new_token_num=}"
            )
        if kv_cache_start + slice_len > kv_cache_token_num:
            raise ValueError(
                f"{kv_cache_start=} + {slice_len=} must be less or equal to {kv_cache_token_num=}"
            )
        if kv_cache_start // page_size != (kv_cache_start + slice_len -
                                           1) // page_size:
            raise ValueError(
                f"Each slice must reside in the same page, but got {kv_cache_start=} and {slice_len=}"
            )

    new_kv_intervals = []
    kv_cache_intervals = []
    for i in range(num_slices[0]):
        new_kv_intervals.append((slices[1][i], slices[1][i] + slices[2][i]))
        kv_cache_intervals.append((slices[0][i], slices[0][i] + slices[2][i]))

    new_kv_intervals.sort()
    kv_cache_intervals.sort()

    # The new_kv slices should be continuous
    for i in range(len(new_kv_intervals) - 1):
        if new_kv_intervals[i][1] != new_kv_intervals[i + 1][0]:
            raise ValueError(
                f"{new_kv_intervals[i][1]=} is expeced to equal to {new_kv_intervals[i + 1][0]}"
            )

    # There should be no overlap among the kv cache slices
    for i in range(len(kv_cache_intervals) - 1):
        if kv_cache_intervals[i][1] > kv_cache_intervals[i + 1][0]:
            raise ValueError(
                f"Overlap detected in kv_cache intervals: {kv_cache_intervals[i]} and {kv_cache_intervals[i+1]}"
            )


def _kv_cache_update(
    new_kv: jax.Array,  # [total_num_token, num_combined_kv_heads, head_dim]
    slices: jax.Array,  # [3, slices], list of (kv_cache_start, new_kv_start,
    # slice_len)
    kv_cache: jax.
    Array,  # [total_num_pages * page_size, num_combined_kv_heads,
    # head_dim]
    num_slices: jax.Array,  # [1]
    page_size: int,
    num_slices_per_block: int,
    dynamic_validate_inputs: bool,
    vmem_limit_bytes: int = 40 * 1024 * 1024,
):
    new_token_num, num_combined_kv_heads, head_dim = new_kv.shape
    assert kv_cache.shape[1] == num_combined_kv_heads
    assert kv_cache.shape[2] == head_dim
    assert head_dim % 128 == 0
    if dynamic_validate_inputs is True:
        _dynamic_validate_inputs(slices, new_token_num, kv_cache.shape[0],
                                 page_size, num_slices)

    in_specs = [
        pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
        pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
    ]

    out_specs = [pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY)]
    out_shape = [jax.ShapeDtypeStruct(kv_cache.shape, dtype=kv_cache.dtype)]

    scalar_prefetches = [slices, num_slices]
    scratch = pltpu.VMEM(
        (num_slices_per_block, page_size, num_combined_kv_heads, head_dim),
        new_kv.dtype,
    )

    scratch_shapes = [
        scratch,
        pltpu.SemaphoreType.DMA,
    ]

    kernel = pl.pallas_call(
        _kv_cache_update_kernel,
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=len(scalar_prefetches),
            in_specs=in_specs,
            out_specs=out_specs,
            grid=(_ceil_div(num_slices[0], num_slices_per_block), ),
            scratch_shapes=scratch_shapes,
        ),
        out_shape=out_shape,
        input_output_aliases={len(scalar_prefetches) + 1: 0},
        compiler_params=pltpu.CompilerParams(
            vmem_limit_bytes=vmem_limit_bytes, ),
    )

    return kernel(*scalar_prefetches, new_kv, kv_cache)[0]


def _prev_power_of_2(n: int) -> int:
    """The previous power of 2 (inclusive)"""
    if n <= 0:
        return 0
    return 1 << (n.bit_length() - 1)


def _get_page_size_bytes(block_size: int, num_combined_kv_heads: int,
                         head_size: int, kv_cache_dtype) -> int:
    """Returns the size in bytes of one page of the KV cache."""
    kv_cache_dtype_bit_size = (dtypes.bit_width(kv_cache_dtype) if hasattr(
        dtypes, "bit_width") else dtypes.itemsize_bits(kv_cache_dtype))
    padded_head_size = _ceil_div(
        head_size, TPU_HEAD_SIZE_ALIGNMENT) * TPU_HEAD_SIZE_ALIGNMENT

    # NOTE: for the implicit padding in XLA
    packing = get_dtype_packing(kv_cache_dtype)
    num_combined_kv_heads = _ceil_div(num_combined_kv_heads, packing) * packing

    return block_size * num_combined_kv_heads * padded_head_size * kv_cache_dtype_bit_size // 8


def _get_num_slices_per_kv_cache_update_block(page_size_bytes: int,
                                              vmem_limit_bytes: int) -> int:
    """Find the optimum number of slices to copy per Pallas program instance.
    Increasing the number of slices copied in one instance of the kernel program
    will increase HBM bandwidth utilization via more in-flight DMAs.
    However, it will also use more VMEM, and experimentally, we observed
    performance regression at 128 slices on v6e, likely due to running
    out of scalar registers. Thus this function will limit the number of
    slices to 64.
    """
    # NOTE: We assume 1MB vmem is used for register spill and others
    assert vmem_limit_bytes >= 1024 * 1024, "vmem_limit_bytes must be at least 1MB"
    num_slices_per_block = (vmem_limit_bytes - 1024 * 1024) // page_size_bytes
    assert num_slices_per_block > 0, "Number of slices should be positive"
    num_slices_per_block = _prev_power_of_2(num_slices_per_block)
    return min(num_slices_per_block, 64)


@functools.partial(
    jax.jit,
    static_argnames=[
        "page_size", "num_slices_per_block", "mesh", "kv_cache_pspec"
    ],
    donate_argnames="kv_cache",
)
def kv_cache_update(
    new_kv: jax.Array,  # [total_num_token, num_combined_kv_heads, head_dim]
    slices: jax.
    Array,  # [3, slices], list of (kv_cache_start, new_kv_start, slice_len)
    kv_cache: jax.
    Array,  # [total_num_pages * page_size, num_combined_kv_heads, head_dim]
    num_slices: jax.Array,  # [1]
    *,
    page_size: int = 32,
    num_slices_per_block: int | None = None,
    mesh: Mesh | None = None,
    kv_cache_pspec: P
    | None = None,  # Only sharding along head_dim is supported
    dynamic_validate_inputs: bool = False,
    vmem_limit_bytes: int = 40 * 1024 * 1024,
):
    if num_slices_per_block is None:
        _, num_combined_kv_heads, head_dim = new_kv.shape
        page_size_bytes = _get_page_size_bytes(page_size,
                                               num_combined_kv_heads, head_dim,
                                               kv_cache.dtype)
        num_slices_per_block = _get_num_slices_per_kv_cache_update_block(
            page_size_bytes, vmem_limit_bytes)

    if mesh is None:
        return _kv_cache_update(new_kv, slices, kv_cache, num_slices,
                                page_size, num_slices_per_block,
                                dynamic_validate_inputs)

    if kv_cache_pspec is None:
        raise ValueError(
            "kv_cache_pspec must be provided when mesh is specified")

    in_specs = (kv_cache_pspec, P(), kv_cache_pspec, P())
    out_specs = kv_cache_pspec
    shard_map_wrapped = jax.shard_map(
        functools.partial(
            _kv_cache_update,
            page_size=page_size,
            num_slices_per_block=num_slices_per_block,
            dynamic_validate_inputs=dynamic_validate_inputs,
            vmem_limit_bytes=vmem_limit_bytes,
        ),
        mesh=mesh,
        in_specs=in_specs,
        out_specs=out_specs,
        check_vma=False,
    )
    return shard_map_wrapped(new_kv, slices, kv_cache, num_slices)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import jax
import jax.numpy as jnp
import torch
from jax.sharding import PartitionSpec
from torch.nn.parameter import Parameter
from torchax.interop import jax_view, torch_view
from torchax.ops.mappings import t2j
from vllm.model_executor.layers.fused_moe import FusedMoE
from vllm.model_executor.layers.linear import LinearBase, LinearMethodBase
from vllm.model_executor.layers.quantization import \
    register_quantization_config
from vllm.model_executor.layers.quantization.awq import (AWQConfig,
                                                         AWQLinearMethod)
from vllm.model_executor.layers.quantization.base_config import \
    QuantizeMethodBase
from vllm.model_executor.layers.quantization.utils.quant_utils import \
    is_layer_skipped

from tpu_inference.layers.common.process_weights.linear_weights import (
    LinearWeights, process_linear_weights, shard_linear_weights,
    to_parameter_list)
from tpu_inference.layers.common.quant_methods import AWQ, get_tpu_quant_method
from tpu_inference.layers.common.quantization import awq_u32_unpack_u4
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation
from tpu_inference.layers.vllm.quantization.configs import (
    VllmQuantConfig, VllmQuantLinearConfig)
from tpu_inference.layers.vllm.quantization.unquantized import \
    VllmUnquantizedLinearMethod
from tpu_inference.logger import init_logger

P = PartitionSpec

logger = init_logger(__name__)


@register_quantization_config(get_tpu_quant_method(AWQ))
class VllmAWQConfig(AWQConfig, VllmQuantConfig):

    @classmethod
    def get_name(cls):
        return AWQ

    def get_supported_act_dtypes(self) -> list[torch.dtype]:
        # NOTE: AWQ checkpoint was quantized with float16. But on TPUs, using
        # bfloat16 is significantly preferred over float16. This might lead to
        # some numeric output change.
        return [torch.bfloat16]

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional[Union["LinearMethodBase", "QuantizeMethodBase"]]:
        if isinstance(layer, LinearBase):
            linear_config = self.get_linear_config(layer)
            if is_layer_skipped(prefix, self.modules_to_not_convert):
                return VllmUnquantizedLinearMethod(linear_config)
            return VllmAWQLinearMethod(self, linear_config)
        elif isinstance(layer, FusedMoE):
            raise NotImplementedError(
                "AWQ FusedMoE is currently not supported in torchax-jax")
        return None


class VllmAWQLinearMethod(AWQLinearMethod):

    def __init__(self, quant_config: VllmAWQConfig,
                 linear_config: VllmQuantLinearConfig):
        super().__init__(quant_config)
        self.linear_config = linear_config

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert layer.qweight.packed_dim == layer.qweight.ndim - 1
        weight = t2j(layer.qweight, use_dlpack=False)
        delattr(layer, "qweight")

        weight_scale = t2j(layer.scales, use_dlpack=False)
        delattr(layer, "scales")

        assert layer.qzeros.packed_dim == layer.qzeros.ndim - 1
        zero_point = t2j(layer.qzeros, use_dlpack=False)
        delattr(layer, "qzeros")

        if layer.bias is not None and not layer.skip_bias_add:
            if layer.return_bias:
                logger.warning_once("Bias might return incorrect value.")
            bias = t2j(layer.bias, use_dlpack=False)
            delattr(layer, "bias")
        else:
            bias = None

        @jax.jit
        def process_awq_linear_weights(
            weight: jax.Array,
            weight_scale: jax.Array,
            zero_point: jax.Array,
            bias: jax.Array | None,
        ) -> LinearWeights:
            weight = awq_u32_unpack_u4(weight)
            group_size = self.quant_config.group_size
            weight = weight.reshape((-1, group_size, weight.shape[-1]))

            zero_point = awq_u32_unpack_u4(zero_point)

            return process_linear_weights(
                LinearWeights(
                    weight=weight,
                    weight_scale=weight_scale,
                    zero_point=zero_point,
                    bias=bias,
                ),
                fused=self.linear_config.fuse_matmuls,
                output_sizes=self.linear_config.output_sizes,
                reorder_size=self.linear_config.n_shards,
                transposed=False,
            )

        weights = process_awq_linear_weights(weight, weight_scale, zero_point,
                                             bias)
        weights = torch_view(
            shard_linear_weights(
                weights,
                mesh=self.linear_config.mesh,
                weight_p_spec=self.linear_config.weight_sharding,
                bias_p_spec=self.linear_config.bias_sharding,
                transposed=False,
            ))

        if self.linear_config.fuse_matmuls:
            layer.qweight = Parameter(weights.weight, requires_grad=False)
            layer.scales = Parameter(weights.weight_scale, requires_grad=False)
            layer.qzeros = Parameter(weights.zero_point, requires_grad=False)
            if bias is not None:
                layer.bias = Parameter(weights.bias, requires_grad=False)
        else:
            layer.qweight = to_parameter_list(weights.weight)
            layer.scales = to_parameter_list(weights.weight_scale)
            layer.qzeros = to_parameter_list(weights.zero_point)
            if bias is not None:
                layer.bias = to_parameter_list(weights.bias)

    def apply(self,
              layer: torch.nn.Module,
              x: torch.Tensor,
              bias: Optional[torch.Tensor] = None) -> torch.Tensor:

        with jax.named_scope(layer._get_name()):
            if self.linear_config.fuse_matmuls:
                out = self._apply_fused(layer, x, bias)
            else:
                out = self._apply_split(layer, x, bias)

        return out

    def _apply_fused(self,
                     layer: torch.nn.Module,
                     x: torch.Tensor,
                     bias: Optional[torch.Tensor] = None) -> torch.Tensor:
        x_jax = jax_view(x)

        qweight = jax_view(layer.qweight)
        qzeros = jnp.expand_dims(jax_view(layer.qzeros), 1)
        scales = jnp.expand_dims(jax_view(layer.scales), 1)

        qweight = qweight.astype(jnp.int8)
        qzeros = qzeros.astype(jnp.int8)

        weight = (qweight - qzeros) * scales
        weight = weight.reshape((-1, weight.shape[-1]))
        outs = jnp.einsum("bd,df->bf", x_jax, weight)

        if bias is not None and not layer.skip_bias_add:
            outs += bias.jax()

        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        out = jnp.concatenate(outs, axis=-1)
        return torch_view(out)

    def _apply_split(self,
                     layer: torch.nn.Module,
                     x: torch.Tensor,
                     bias: Optional[torch.Tensor] = None) -> torch.Tensor:
        assert isinstance(layer.qweight, torch.nn.ParameterList)

        x_jax = jax_view(x)
        params = zip(layer.qweight, layer.qzeros, layer.scales)
        outs = []
        for i, (qweight, qzeros, scales) in enumerate(params):
            qweight = jax_view(qweight)
            scales = jnp.expand_dims(jax_view(scales), 1)
            qzeros = jnp.expand_dims(jax_view(qzeros), 1)

            qweight = qweight.astype(jnp.int8)
            qzeros = qzeros.astype(jnp.int8)

            weight = (qweight - qzeros) * scales
            weight = weight.reshape((-1, weight.shape[-1]))
            out = jnp.einsum("bd,df->bf", x_jax, weight)

            if bias is not None and not layer.skip_bias_add:
                out += jax_view(bias[i])

            outs.append(out)
        out = jnp.concatenate(outs, axis=-1)
        return torch_view(out)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from vllm.attention.layer import Attention
from vllm.config import (CacheConfig, ModelConfig, ParallelConfig,
                         SchedulerConfig, VllmConfig)
from vllm.sampling_params import SamplingType
from vllm.v1.attention.backend import AttentionType
from vllm.v1.kv_cache_interface import (FullAttentionSpec, KVCacheConfig,
                                        KVCacheGroupSpec, KVCacheTensor,
                                        MLAAttentionSpec, SlidingWindowSpec)
from vllm.v1.request import Request

from tpu_inference import utils as common_utils
from tpu_inference.runner.input_batch import CachedRequestState
from tpu_inference.runner.kv_cache import get_attention_page_size_bytes
from tpu_inference.runner.tpu_runner import TPUModelRunner


class TestKVCacheManager:

    def _setup_runner(self, use_mla: bool = False):
        # Mock JAX dependencies
        self.mock_rng_key = MagicMock()

        self.mock_devices = [MagicMock(coords=i) for i in range(4)]
        self.mock_rng_key = MagicMock()

        # create 1x1 mesh
        devices = np.asarray(jax.devices()[:1])
        axis_names = ('data', 'attn_dp', 'model', 'expert')
        mesh_shape = (1, 1, 1, 1)
        self.mock_mesh = jax.sharding.Mesh(devices.reshape(mesh_shape),
                                           axis_names)

        with patch('jax.devices', return_value=self.mock_devices), \
             patch('jax.make_mesh', return_value=self.mock_mesh), \
             patch('jax.experimental.mesh_utils.create_device_mesh', return_value=self.mock_mesh), \
             patch('tpu_inference.runner.tpu_runner.TPUModelRunner._create_new_model_mesh', return_value=self.mock_mesh), \
             patch('tpu_inference.runner.tpu_runner.TPUModelRunner._init_mesh', return_value=self.mock_mesh), \
             patch('jax.random.key', return_value=self.mock_rng_key), \
             patch('tpu_inference.runner.tpu_runner.get_model', return_value=MagicMock()):

            model_config = ModelConfig(tokenizer_mode="auto",
                                       trust_remote_code=False,
                                       seed=0,
                                       dtype='bfloat16',
                                       use_mla=use_mla)
            cache_config = CacheConfig(
                block_size=16,
                gpu_memory_utilization=0.9,
                swap_space=4,
                cache_dtype="auto",
            )
            scheduler_config = SchedulerConfig(max_num_seqs=16,
                                               max_model_len=1024,
                                               is_encoder_decoder=False)
            parallel_config = ParallelConfig(
                pipeline_parallel_size=1,
                tensor_parallel_size=1,
                worker_use_ray=False,
            )
            vllm_config = VllmConfig(
                model_config=model_config,
                cache_config=cache_config,
                scheduler_config=scheduler_config,
                parallel_config=parallel_config,
                observability_config={},
                additional_config={},
            )
            self.runner = TPUModelRunner(vllm_config,
                                         devices=self.mock_devices)
            self.runner.mesh = self.mock_mesh

    def setup_method(self):
        self._setup_runner(use_mla=False)

    def test_insert_request_with_kv_cache(self):
        # This test refines the insertion test by first extracting a KV cache
        # using get_kv_cache_for_block_ids, simulating a prefill->decode
        # transfer, and then inserting it. This ensures the extraction and
        # insertion logic are compatible.

        # 1. ===== Setup source runner for prefill simulation =====
        self.runner.block_size = 64
        num_layers = 2
        num_kv_heads = 16
        head_size = 128
        num_blocks = 50
        # This is needed for the padding logic in insert_request_with_kv_cache
        self.runner.vllm_config.cache_config.num_gpu_blocks = num_blocks

        prompt_len = 64

        # Populate a source KV cache with data. This represents the state
        # of the prefill runner's KV cache.
        source_kv_cache_shape = (num_blocks, self.runner.block_size,
                                 2 * num_kv_heads // 2, 2, head_size)
        prod_val = int(np.prod(source_kv_cache_shape))
        source_kv_caches = [
            jnp.arange(prod_val,
                       dtype=jnp.bfloat16).reshape(source_kv_cache_shape),
            jnp.arange(prod_val, 2 * prod_val,
                       dtype=jnp.bfloat16).reshape(source_kv_cache_shape)
        ]
        self.runner.kv_caches = source_kv_caches

        # Create a mock for sampling_params to avoid TypeErrors in add_request
        mock_sampling_params = MagicMock()
        mock_sampling_params.sampling_type = SamplingType.GREEDY
        mock_sampling_params.temperature = 0.0
        mock_sampling_params.top_p = 1.0
        mock_sampling_params.top_k = -1  # Common value for greedy
        mock_sampling_params.min_tokens = 0
        mock_sampling_params.logprobs = None
        mock_sampling_params.logit_bias = None
        mock_sampling_params.allowed_token_ids = set()
        mock_sampling_params.bad_words_token_ids = None
        mock_sampling_params.all_stop_token_ids = set()

        # 2. ===== Simulate prefill execution state =====
        prefill_block_ids = [5]
        # Create a request state for prefill.
        prefill_request_state = CachedRequestState(
            req_id="test_req_1",
            prompt_token_ids=list(range(prompt_len)),
            output_token_ids=[],
            sampling_params=mock_sampling_params,
            block_ids=tuple([prefill_block_ids]),
            num_computed_tokens=0,
            lora_request=None,
            mm_features=[],
            pooling_params=None,
            generator=None,
        )

        # Add the request to the input_batch to simulate it being scheduled.
        self.runner.input_batch.add_request(prefill_request_state)

        # 3. ===== Extract KV cache using get_kv_cache_for_block_ids =====
        # Extract the full KV cache for the allocated block.
        full_block_kv_cache = self.runner.get_kv_cache_for_block_ids(
            block_ids=prefill_block_ids)

        # Since get_kv_cache_for_block_ids returns the full block, but the
        # prompt only fills part of it, we need to slice it to the actual
        # prompt length for the insertion test to be accurate.
        extracted_kv_cache_slices = [
            layer_cache[:prompt_len] for layer_cache in full_block_kv_cache
        ]

        # 4. ===== Setup destination runner for decode simulation =====
        # Reset runner state to simulate a fresh decode runner.
        self.runner.requests = {}
        req_index = self.runner.input_batch.remove_request("test_req_1")
        if req_index is not None:
            self.runner.input_batch.condense([req_index])

        # Initialize destination KV caches with zeros.
        dest_kv_cache_shape = (num_blocks, self.runner.block_size,
                               2 * num_kv_heads // 2, 2, head_size)
        self.runner.kv_caches = [
            jnp.zeros(dest_kv_cache_shape, dtype=jnp.bfloat16)
            for _ in range(num_layers)
        ]

        # Create a mock request as it would be after prefill + 1 token.
        decode_request = MagicMock(spec=Request)
        decode_request.request_id = "test_req_1"
        decode_request.num_tokens = prompt_len + 1  # Total tokens
        decode_request.num_computed_tokens = prompt_len
        decode_request.prompt_token_ids = list(range(prompt_len))
        decode_request.all_token_ids = [123, 232, 908]
        decode_request.output_token_ids = [100]
        decode_request.sampling_params = mock_sampling_params

        decode_request.lora_request = None
        decode_request.mm_kwargs, decode_request.mm_positions = [], []
        decode_request.pooling_params, decode_request.generator = None, None

        # Prepare the KV cache slices for insertion. They must be padded to the
        # full block size and have a leading dimension for the number of blocks.

        # Allocate new block IDs for the decode runner.
        decode_block_ids = [[10]]
        # 5. ===== Call the method to be tested =====
        self.runner.insert_request_with_kv_cache(decode_request,
                                                 extracted_kv_cache_slices,
                                                 decode_block_ids)

        # 6. ===== Assertions =====
        assert "test_req_1" in self.runner.requests
        assert "test_req_1" in self.runner.input_batch.req_id_to_index
        assert self.runner.requests[
            "test_req_1"].num_computed_tokens == prompt_len
        assert self.runner.requests["test_req_1"].output_token_ids == [908]

        # Verify the content of the inserted KV cache.
        target_block_id = decode_block_ids[0][0]
        for i, layer_kv_cache in enumerate(self.runner.kv_caches):
            updated_block_content = layer_kv_cache[target_block_id]

            # The extracted slice should be padded to the block size.
            padding_size = self.runner.block_size - prompt_len
            expected_padded_slice = jnp.pad(extracted_kv_cache_slices[i],
                                            ((0, padding_size), (0, 0), (0, 0),
                                             (0, 0)),
                                            mode='constant')
            np.testing.assert_array_equal(updated_block_content,
                                          expected_padded_slice)

    @pytest.mark.parametrize("num_kv_heads", [16, 32])
    @pytest.mark.parametrize("head_size", [64, 100, 200])
    def test_get_kv_cache_spec_with_compilation_cfg(self, num_kv_heads,
                                                    head_size):
        # tests we create kv cache spec from compilation config
        # create a static forward context with
        # 10 full attention layers +
        # 10 sliding window attention layers
        # 1 layer with shared kv cache.
        attn_type = AttentionType.DECODER
        sliding_window = 10
        static_forward_context = {}
        for i in range(10):
            static_forward_context[f'layer.{i}'] = MagicMock(
                spec=Attention,
                num_kv_heads=num_kv_heads,
                head_size=head_size,
                attn_type=attn_type,
                sliding_window=None,
                kv_sharing_target_layer_name=None,
            )
        for i in range(10, 20):
            static_forward_context[f'layer.{i}'] = MagicMock(
                spec=Attention,
                num_kv_heads=num_kv_heads,
                head_size=head_size,
                attn_type=attn_type,
                sliding_window=sliding_window,
                kv_sharing_target_layer_name=None,
            )
        static_forward_context['layer.20'] = MagicMock(
            spec=Attention,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            attn_type=attn_type,
            sliding_window=None,
            kv_sharing_target_layer_name='layer.0',
        )
        self.runner.vllm_config.compilation_config.static_forward_context = \
            static_forward_context

        kv_cache_spec = self.runner.get_kv_cache_spec()

        block_size = self.runner.vllm_config.cache_config.block_size
        num_kv_heads = common_utils.get_padded_num_heads(
            num_kv_heads, self.runner.mesh.shape["model"])
        head_size = common_utils.get_padded_head_dim(head_size)

        expected_full_attn_spec = FullAttentionSpec(
            block_size=block_size,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            dtype=torch.bfloat16,
            page_size_padded=get_attention_page_size_bytes(
                self.runner.mesh, block_size, num_kv_heads, head_size,
                self.runner.kv_cache_dtype, False))
        expected_sliding_window_spec = SlidingWindowSpec(
            block_size=block_size,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            dtype=torch.bfloat16,
            sliding_window=sliding_window,
            page_size_padded=get_attention_page_size_bytes(
                self.runner.mesh, block_size, num_kv_heads, head_size,
                self.runner.kv_cache_dtype, False))
        assert len(kv_cache_spec) == 20
        for i in range(10):
            assert kv_cache_spec[f'layer.{i}'] == expected_full_attn_spec
        for i in range(10, 20):
            assert kv_cache_spec[f'layer.{i}'] == expected_sliding_window_spec
        assert 'layer.20' not in kv_cache_spec
        assert self.runner.kv_cache_manager.shared_kv_cache_layers == {
            'layer.20': 'layer.0'
        }

    def test_get_kv_cache_spec_with_compilation_cfg_mla(self):
        # tests we create kv cache spec from compilation config with mla
        self.runner.kv_cache_manager.use_mla = True

        # Mock hf_text_config to have kv_lora_rank and qk_rope_head_dim
        mock_hf_text_config = MagicMock()
        mock_hf_text_config.kv_lora_rank = 400
        mock_hf_text_config.qk_rope_head_dim = 40
        self.runner.model_config.hf_text_config = mock_hf_text_config

        num_kv_heads = 16
        head_size = 512  # Aggregated padding amount may be passed to the model instead.
        expected_head_size = 640  # 640 = align(512, 128) + alignto(40, 128)
        attn_type = AttentionType.DECODER
        static_forward_context = {}
        # Mock one layer, as the logic is the same for all
        mock_attn_module = MagicMock(
            spec=Attention,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            attn_type=attn_type,
            sliding_window=None,
            kv_sharing_target_layer_name=None,
        )
        mock_attn_module.use_mla = True
        static_forward_context['layer.0'] = mock_attn_module
        self.runner.vllm_config.compilation_config.static_forward_context = \
            static_forward_context

        kv_cache_spec = self.runner.get_kv_cache_spec()

        assert len(kv_cache_spec) == 1
        spec = kv_cache_spec['layer.0']
        assert isinstance(spec, MLAAttentionSpec)
        assert spec.num_kv_heads == 1
        assert spec.head_size == expected_head_size

    def test_get_kv_cache_spec_without_compilation_cfg(self):
        # tests if there's no compilation config, we use full attention kv
        # cache for each layer.
        model_config = self.runner.vllm_config.model_config
        parallel_config = self.runner.vllm_config.parallel_config
        head_size = model_config.get_head_size()
        num_kv_heads = model_config.get_total_num_kv_heads()
        num_layers = model_config.get_num_layers(parallel_config)

        self.runner.vllm_config.compilation_config.static_forward_context = {}
        kv_cache_spec = self.runner.get_kv_cache_spec()

        assert len(kv_cache_spec) == num_layers
        block_size = self.runner.vllm_config.cache_config.block_size
        num_kv_heads = common_utils.get_padded_num_heads(
            num_kv_heads, self.runner.mesh.shape["model"])
        head_size = common_utils.get_padded_head_dim(head_size)
        expected_full_attn_spec = FullAttentionSpec(
            block_size=block_size,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            dtype=torch.bfloat16,
            page_size_padded=get_attention_page_size_bytes(
                self.runner.mesh, block_size, num_kv_heads, head_size,
                self.runner.kv_cache_dtype, False))
        for i in range(num_layers):
            assert kv_cache_spec[f'layer.{i}'] == expected_full_attn_spec
        assert len(self.runner.kv_cache_manager.shared_kv_cache_layers) == 0

    def test_get_kv_cache_spec_without_compilation_cfg_mla(self):
        self.runner.kv_cache_manager.use_mla = True
        model_config = self.runner.vllm_config.model_config
        parallel_config = self.runner.vllm_config.parallel_config
        num_layers = model_config.get_num_layers(parallel_config)

        mock_hf_text_config = MagicMock()
        mock_hf_text_config.kv_lora_rank = 400
        mock_hf_text_config.qk_rope_head_dim = 40
        self.runner.model_config.hf_text_config = mock_hf_text_config
        expected_head_size = 640  # 640 = align(512, 128) + alignto(40, 128)

        self.runner.vllm_config.compilation_config.static_forward_context = {}
        with patch('vllm.config.ModelConfig.get_num_layers',
                   return_value=num_layers):
            kv_cache_spec = self.runner.get_kv_cache_spec()

        assert len(kv_cache_spec) == num_layers
        for i in range(num_layers):
            spec = kv_cache_spec[f"layer.{i}"]
            assert isinstance(spec, MLAAttentionSpec)
            assert spec.num_kv_heads == 1
            assert spec.head_size == expected_head_size

    def test_initialize_kv_cache(self):
        # create a kv cache config with 10 layers full attention and 10 layers
        # sliding window attention.
        block_size = self.runner.vllm_config.cache_config.block_size
        num_kv_heads = 8
        head_size = 128
        sliding_window = 100
        num_blocks = 100
        kv_packing = 2  #bf16
        sliding_window_spec = SlidingWindowSpec(
            block_size=block_size,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            dtype=torch.bfloat16,
            sliding_window=sliding_window,
        )
        full_attn_spec = FullAttentionSpec(
            block_size=block_size,
            num_kv_heads=num_kv_heads,
            head_size=head_size,
            dtype=torch.bfloat16,
        )
        kv_cache_groups = [
            KVCacheGroupSpec(layer_names=[f'layer.{i}' for i in range(10)],
                             kv_cache_spec=full_attn_spec),
            KVCacheGroupSpec(layer_names=[f'layer.{i}' for i in range(10, 20)],
                             kv_cache_spec=sliding_window_spec),
        ]
        kv_cache_tensors = []
        page_size_bytes = full_attn_spec.page_size_bytes
        for i in range(10):
            kv_cache_tensors.append(
                KVCacheTensor(
                    size=num_blocks * page_size_bytes,
                    shared_by=[f'layer.{i}', f'layer.{i+10}'],
                ))
        kv_cache_config = KVCacheConfig(
            num_blocks=num_blocks,
            kv_cache_tensors=kv_cache_tensors,
            kv_cache_groups=kv_cache_groups,
        )

        original_input_batch = self.runner.input_batch
        self.runner.initialize_kv_cache(kv_cache_config)

        # assert kv cache config with multiple kv cache groups will reinit
        # input batch.
        assert original_input_batch != self.runner.input_batch
        assert len(self.runner.kv_caches) == 10
        for i in range(10):
            assert self.runner.kv_caches[i].shape == (num_blocks, block_size,
                                                      num_kv_heads * 2 //
                                                      kv_packing, kv_packing,
                                                      head_size)
            assert self.runner.layer_name_to_kvcache_index[f'layer.{i}'] == i
            assert self.runner.layer_name_to_kvcache_index[
                f'layer.{i + 10}'] == i

    def test_get_kv_cache_spec_with_eagle3(self):
        # tests we create kv cache spec for eagle3 draft model
        self.runner.vllm_config.compilation_config.static_forward_context = {}
        mock_speculative_config = MagicMock()
        mock_speculative_config.method = "eagle3"
        mock_draft_model_config = MagicMock()
        mock_hf_config = MagicMock()
        mock_hf_config.num_key_value_heads = 4
        mock_hf_config.hidden_size = 1024
        mock_hf_config.num_attention_heads = 8
        mock_draft_model_config.hf_config = mock_hf_config
        mock_speculative_config.draft_model_config = mock_draft_model_config
        self.runner.speculative_config = mock_speculative_config

        kv_cache_spec = self.runner.get_kv_cache_spec()

        assert "draft_layer.0" in kv_cache_spec
        draft_spec = kv_cache_spec["draft_layer.0"]
        assert isinstance(draft_spec, FullAttentionSpec)
        assert draft_spec.block_size == self.runner.vllm_config.cache_config.block_size
        assert draft_spec.num_kv_heads == common_utils.get_padded_num_heads(
            4, self.runner.mesh.shape["model"])
        assert draft_spec.head_size == common_utils.get_padded_head_dim(128)
        assert draft_spec.dtype == torch.bfloat16

    def test_get_kv_cache_spec_with_eagle3_mla(self):
        # tests we create kv cache spec for eagle3 draft model with mla
        self.runner.kv_cache_manager.use_mla = True

        self.runner.vllm_config.compilation_config.static_forward_context = {}
        mock_speculative_config = MagicMock()
        mock_speculative_config.method = "eagle3"
        mock_draft_model_config = MagicMock()
        mock_hf_config = MagicMock()
        mock_hf_config.num_key_value_heads = 4
        mock_hf_config.hidden_size = 1024
        mock_hf_config.num_attention_heads = 8
        mock_hf_config.num_layers = 16
        model_layers = 1
        mock_hf_text_config = MagicMock()
        mock_hf_text_config.kv_lora_rank = 400
        mock_hf_text_config.qk_rope_head_dim = 40
        self.runner.model_config.hf_text_config = mock_hf_text_config
        mock_draft_model_config.hf_config = mock_hf_config
        mock_speculative_config.draft_model_config = mock_draft_model_config
        self.runner.speculative_config = mock_speculative_config

        kv_cache_spec = self.runner.get_kv_cache_spec()

        # Without compilation context, it will create specs for the main model layers
        # as well as the draft model layer.
        assert len(kv_cache_spec) > model_layers

        assert "draft_layer.0" in kv_cache_spec
        draft_spec = kv_cache_spec["draft_layer.0"]
        assert isinstance(draft_spec, FullAttentionSpec)

        for i in range(model_layers):
            assert f"layer.{i}" in kv_cache_spec
            spec = kv_cache_spec[f"layer.{i}"]
            assert isinstance(spec, MLAAttentionSpec)
            assert spec.num_kv_heads == 1



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

import jax
import jax.numpy as jnp
from jax.experimental import xla_metadata
from jax.experimental.pallas.ops.tpu.megablox.gmm import gmm as megablox_gmm
from qwix._src.core.ragged_dot import ragged_dot as qwix_ragged_dot

from tpu_inference import envs
from tpu_inference.layers.common.fused_moe_gmm import \
    round_up_to_multiple_of_128_within_limit
from tpu_inference.layers.jax.layers import FlaxUtils
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.utils.qwix.qwix_utils import \
    manually_quantize_qwix_activation

logger = init_logger(__name__)
modeling_flax_utils = FlaxUtils()
set_xla_metadata = xla_metadata.set_xla_metadata


class MoEBackend(enum.Enum):
    FUSED_MOE = "fused_moe"
    VLLM_MOE = "vllm_moe"
    DENSE_MAT = "dense_mat"
    MEGABLX_GMM = "megablox_gmm"
    RAGGED_DOT = "ragged_dot_gmm"


def select_moe_backend():
    # Validation: Ensure mutually exclusive flags aren't set together
    if envs.USE_MOE_EP_KERNEL and envs.USE_VLLM_MOE_KERNEL:
        raise ValueError("Cannot enable multiple MoE kernels simultaneously.")

    if envs.USE_MOE_EP_KERNEL:
        logger.info("[MoE]: Fused MoE kernel is enabled")
        return MoEBackend.FUSED_MOE

    if envs.USE_VLLM_MOE_KERNEL:
        logger.info("[MoE]: VLLM MoE kernel is enabled")
        return MoEBackend.VLLM_MOE

    if envs.USE_MEGABLOCKS:
        logger.info("[MoE]: Mega Blocks is enabled for GMM in Sparse Matmul")
        return MoEBackend.MEGABLX_GMM

    if envs.USE_RAGGED_DOT:
        logger.info("[MoE]: Ragged Dot is enabled for GMM in Sparse Matmul")
        return MoEBackend.RAGGED_DOT

    # Default case
    logger.info("[MoE]: Dense Matmul is enabled")
    return MoEBackend.DENSE_MAT


# --- Helper Functions/Class for Sparse MoE ---
class TransformStrategy(enum.Enum):
    INPUT_OFFSET = enum.auto()
    SEND_SIZE = enum.auto()
    OUTPUT_OFFSET = enum.auto()
    RECV_SIZE = enum.auto()


def sort_activations_fn(inputs: jax.Array,
                        sort_indices: jax.Array) -> jax.Array:
    """Stateless sort of activations."""
    return inputs[sort_indices, ...]


def global_permute_fn(inputs_TD: jax.Array, selected_experts_TX: jax.Array,
                      num_experts_per_tok: int, num_local_experts: int):
    """Stateless global permute: Sorts tokens by assigned expert."""
    total_tokens = inputs_TD.shape[0]
    flat_expert_indices = selected_experts_TX.flatten()
    sort_indices_t = jnp.argsort(flat_expert_indices)

    replicated_inputs_tD = jnp.repeat(inputs_TD, num_experts_per_tok, axis=0)
    sorted_inputs_tD = sort_activations_fn(replicated_inputs_tD,
                                           sort_indices_t)

    # number of tokens assigned to each expert
    group_sizes_E = jnp.bincount(flat_expert_indices, length=num_local_experts)

    expert_ids = jnp.arange(num_local_experts)
    total_assignments = total_tokens * num_experts_per_tok
    sorted_expert_assignments_t = jnp.repeat(
        expert_ids,
        repeats=group_sizes_E,
        total_repeat_length=total_assignments)

    return (
        sorted_inputs_tD,
        sort_indices_t,
        group_sizes_E,
        sorted_expert_assignments_t,
    )


def unpermute_fn(processed_tokens: jax.Array, sort_indices: jax.Array,
                 router_weights_TX: jax.Array, num_experts_per_tok: int,
                 hidden_size: int, output_dtype):
    """Stateless global unpermute logic."""
    with jax.named_scope("unpermute"):
        unsorted_tokens_tD = sort_activations_fn(processed_tokens,
                                                 jnp.argsort(sort_indices))
        reshaped_tokens_TXD = unsorted_tokens_tD.reshape(
            -1, num_experts_per_tok, hidden_size)

    with jax.named_scope("combine_weights"):
        output_TD = jnp.einsum(
            "TXD,TX -> TD",
            reshaped_tokens_TXD.astype(jnp.float32),
            router_weights_TX.astype(jnp.float32),
            precision='float32',
        )

    return output_TD.astype(output_dtype)


def local_permute_fn(inputs,
                     global_group_sizes,
                     local_expert_size,
                     shard_index,
                     is_offset,
                     global_sorted_experts=None):
    """Stateless local permutation logic."""
    # global_group_sizes: (tokens parallelism, num_total_experts)
    # all_shard_local_sizes: (tokens parallelism, num local experts in the shard)
    start_index = shard_index * local_expert_size
    all_shard_local_sizes = jax.lax.dynamic_slice_in_dim(global_group_sizes,
                                                         start_index,
                                                         local_expert_size,
                                                         axis=1)
    local_sizes = all_shard_local_sizes.reshape(-1)

    # local_group_size: (tokens parallelism, )
    local_group_size = jnp.sum(all_shard_local_sizes, axis=0)

    # When token replicated in devices
    if is_offset:
        global_sorted_shard_assignments = jnp.floor_divide(
            global_sorted_experts, local_expert_size)
        expert_indices = jnp.where(
            global_sorted_shard_assignments == shard_index,
            jnp.mod(global_sorted_experts, local_expert_size),
            local_expert_size)

    # When token sharded in devices
    else:
        base_indices = jnp.mod(jnp.arange(local_sizes.shape[0]),
                               local_expert_size)
        expert_indices = jnp.repeat(base_indices,
                                    local_sizes,
                                    total_repeat_length=inputs.shape[0])

    sorted_indices = jnp.argsort(expert_indices)
    # sort the inputs based on the local expert_indices
    sorted_inputs = sort_activations_fn(inputs, sorted_indices)
    # sorted local expert id from 0 to local expert size
    sorted_experts_ids = expert_indices[sorted_indices]

    return (
        sorted_inputs,
        sorted_indices,
        local_group_size,
        sorted_experts_ids,
    )


def get_all_to_all_params_fn(all_shards_group_sizes,
                             shard_id,
                             num_expert_parallelism,
                             is_batch_sharded=True):
    """Stateless parameter generation for ragged_all_to_all."""

    def transform_array(input_array, shard_id, strategy, is_batch_sharded):
        if is_batch_sharded:
            if strategy == TransformStrategy.INPUT_OFFSET:
                local_array = input_array[shard_id]
                return jnp.concatenate(
                    (jnp.array([0]), jnp.cumsum(local_array)[:-1]))
            elif strategy == TransformStrategy.SEND_SIZE:
                return input_array[shard_id]
            elif strategy == TransformStrategy.OUTPUT_OFFSET:
                zero_row = jnp.zeros((1, ) + input_array.shape[1:],
                                     dtype=input_array.dtype)
                array_with_zeros = jnp.concatenate((zero_row, input_array),
                                                   axis=0)
                cumulated_array = jnp.cumsum(array_with_zeros,
                                             axis=0,
                                             dtype=input_array.dtype)
                return cumulated_array[shard_id]
            elif strategy == TransformStrategy.RECV_SIZE:
                return input_array[:, shard_id]
            else:
                raise ValueError(
                    f"Unknown transform array strategy: {strategy}")
        else:
            if strategy == TransformStrategy.INPUT_OFFSET:
                return jnp.zeros(num_expert_parallelism,
                                 dtype=input_array.dtype)
            elif strategy == TransformStrategy.SEND_SIZE:
                return jnp.repeat(input_array[shard_id],
                                  num_expert_parallelism)
            elif strategy == TransformStrategy.OUTPUT_OFFSET:
                output_offset = jnp.concatenate(
                    (jnp.array([0]), jnp.cumsum(input_array[:-1])))[shard_id]
                return jnp.repeat(output_offset, num_expert_parallelism)
            elif strategy == TransformStrategy.RECV_SIZE:
                return input_array
            else:
                raise ValueError(
                    f"Unknown transform array strategy: {strategy}")

    input_offsets = transform_array(all_shards_group_sizes, shard_id,
                                    TransformStrategy.INPUT_OFFSET,
                                    is_batch_sharded)
    send_sizes = transform_array(all_shards_group_sizes, shard_id,
                                 TransformStrategy.SEND_SIZE, is_batch_sharded)
    output_offsets = transform_array(all_shards_group_sizes, shard_id,
                                     TransformStrategy.OUTPUT_OFFSET,
                                     is_batch_sharded)
    recv_sizes = transform_array(all_shards_group_sizes, shard_id,
                                 TransformStrategy.RECV_SIZE, is_batch_sharded)
    return input_offsets, send_sizes, output_offsets, recv_sizes


def gmm_fn(inputs, kernel, group_sizes, tile_size, moe_backend, dtype,
           quantized_dtype):
    """Stateless Grouped Matrix Multiply."""
    num_rows = inputs.shape[0]
    pad_amount = (tile_size[0] - num_rows % tile_size[0]) % tile_size[0]
    if pad_amount > 0:
        inputs = jnp.pad(inputs, ((0, pad_amount), (0, 0)))

    if moe_backend == MoEBackend.MEGABLX_GMM:
        m = inputs.shape[0]
        _, k, n = kernel.shape
        tm = round_up_to_multiple_of_128_within_limit(m, 512)
        tk = round_up_to_multiple_of_128_within_limit(k, 2048)
        tn = round_up_to_multiple_of_128_within_limit(n, 2048)

        output = megablox_gmm(lhs=inputs,
                              rhs=kernel,
                              group_sizes=group_sizes,
                              preferred_element_type=dtype,
                              tiling=(tm, tk, tn))

    elif moe_backend == MoEBackend.RAGGED_DOT:
        inputs = manually_quantize_qwix_activation(
            inputs, "ragged_dot", jnp.float8_e4m3fn, [0], {},
            "absmax") if quantized_dtype else inputs
        ragged_dot_func = qwix_ragged_dot if quantized_dtype else jax.lax.ragged_dot
        with set_xla_metadata(ragged_dot_tiling=",".join(
            [str(t) for t in tile_size]),
                              mosaic_fusion_group="ragged-dot"):
            output = ragged_dot_func(lhs=inputs,
                                     rhs=kernel,
                                     group_sizes=group_sizes,
                                     preferred_element_type=dtype)

    if pad_amount > 0:
        output = output[:num_rows, :]
    return output



# SPDX-License-Identifier: Apache-2.0
"""Quantized matmul kernel with blockwise quantization support."""

import functools

import jax
import jax.numpy as jnp
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

from tpu_inference.kernels.quantized_matmul import util
from tpu_inference.kernels.quantized_matmul.tuned_block_sizes import (
    TunedValue, get_device_vmem_limit, get_tuned_block_sizes)
from tpu_inference.kernels.quantized_matmul.util import (get_kernel_name,
                                                         next_multiple,
                                                         unfold_args)

quantize_tensor = util.quantize_tensor
MXU_SIZE = 256


@functools.partial(
    jax.jit,
    static_argnames=[
        "block_size",
        "x_q_dtype",
        "tuned_value",
    ],
)
def quantized_matmul_kernel(
    x: jax.Array,  # [bs, n_in]
    w_q: jax.Array,  # [n_out, n_in]
    w_scale: jax.Array,  # [n_in // block_size, 1, n_out]
    w_zp: jax.Array | None = None,  # [n_out]
    block_size: int | None = None,
    x_q_dtype: jnp.dtype | None = None,
    *,
    tuned_value: TunedValue | None = None,
) -> jax.Array:
    """Quantized matmul kernel with blockwise support.

    Args:
      x: Input unquantized array.
      w_q: Weight quantized array. [n_output_features, n_input_features]
      w_scale: Weight quantization scale. [n_input_features // block_size, 1, n_output_features]
      w_zp: Weight zero point for asymmetric quantization.
      block_size: Block size for subchannel quantization.
      x_q_dtype: Quantization type of the input. If None or if the value is the
        same as x.dtype, then no quantization is applied.
      tuned_value: Kernel tuned values for optimal performance.

    Returns:
      Quantized matmul result.
    """

    if w_zp is not None:
        raise NotImplementedError("zero_point is not supported.")

    if x_q_dtype is None:
        x_q_dtype = x.dtype
    quantize_activation = x_q_dtype != x.dtype

    orig_n_batch, orig_n_in = x.shape
    orig_n_out, *_ = w_q.shape

    if tuned_value is None:
        tuned_value = get_tuned_block_sizes(
            n_batch=orig_n_batch,
            n_out=orig_n_out,
            n_in=orig_n_in,
            x_q_dtype=jnp.dtype(x_q_dtype).name,
            w_q_dtype=jnp.dtype(w_q.dtype).name,
        )
    batch_block_size = tuned_value.batch_block_size
    out_block_size = tuned_value.out_block_size
    in_block_size = tuned_value.in_block_size
    n_lane_multiplier = tuned_value.n_lane_multiplier
    block_size = tuned_value.in_block_size if block_size is None else block_size

    # Pad the inputs to be multiple of block size.
    padded_n_batch = next_multiple(orig_n_batch, batch_block_size)
    if orig_n_batch < padded_n_batch:
        x = jnp.pad(x, ((0, padded_n_batch - orig_n_batch), (0, 0)))

    padded_n_out = next_multiple(orig_n_out, out_block_size)
    if orig_n_out < padded_n_out:
        w_q = jnp.pad(w_q, ((0, padded_n_out - orig_n_out), (0, 0)))
        w_scale = jnp.pad(w_scale, (0, padded_n_out - orig_n_out))
    padded_n_in = next_multiple(orig_n_in, in_block_size)
    if orig_n_in < padded_n_in:
        x = jnp.pad(x, ((0, 0), (0, padded_n_in - orig_n_in)))
        w_q = jnp.pad(w_q, ((0, 0), (0, padded_n_in - orig_n_in)))

    if w_scale.dtype != jnp.float32:
        w_scale = w_scale.astype(jnp.float32)

    n_batch = padded_n_batch // batch_block_size
    n_out = padded_n_out // out_block_size
    n_in = padded_n_in // in_block_size

    save_acc = n_in > 1
    # Remove redundant input quantization logic by caching quantized input. For
    # best performance, only enable this behavior when single input block is
    # used per batch.
    save_x_q = quantize_activation and n_in == 1 and n_out > 1

    # TODO(amandaliang): Make this configurable.
    acc_dtype = jnp.bfloat16
    if quantize_activation and jnp.issubdtype(w_q.dtype, jnp.integer):
        acc_dtype = jnp.int32

    vmem_limit_bytes = util.get_vmem_limit(
        n_batch=n_batch,
        n_out=n_out,
        n_in=n_in,
        batch_block_size=batch_block_size,
        out_block_size=out_block_size,
        in_block_size=in_block_size,
        x_dtype=x.dtype,
        x_q_dtype=x_q_dtype,
        w_q_dtype=w_q.dtype,
        scale_dtype=jnp.float32,
        out_dtype=x.dtype,
        acc_dtype=acc_dtype,
        save_acc=save_acc,
        save_x_q=save_x_q,
        upper_limit_bytes=get_device_vmem_limit(),
    )

    steps_k = in_block_size // block_size
    # n_lane_multiplier > 1 could improve perf by reducing loop overhead and increasing instruction-level parallelism,
    # allowing the compiler to overlap output fusion and packing overhead with MXU computation
    # TODO(amandaliang): use pltpu.get_tpu_info().mxu_column_size when JAX version is newer
    compute_tile_n = MXU_SIZE * n_lane_multiplier
    steps_n = out_block_size // compute_tile_n

    def kernel(lhs_ref, rhs_ref, w_scales_ref, out_ref, acc_scratch):
        pid_k = pl.program_id(2)
        is_first_step = pid_k == 0
        is_last_step = pid_k == (orig_n_in // in_block_size - 1)

        def accum(is_first_step, is_last_step):
            accumulators = [None] * steps_n

            for i in range(steps_k):
                k_start, k_end = i * block_size, (i + 1) * block_size
                lhs_sub = lhs_ref[:, k_start:k_end].astype(jnp.float32)
                lhs_q, lhs_scale = util.quantize_block(lhs_sub, 1, x_q_dtype)
                lhs_scale = lhs_scale.astype(acc_dtype)

                rhs_q_full = rhs_ref[:, k_start:k_end]
                rhs_scale_full = w_scales_ref[i, :, :].astype(acc_dtype)

                for j in range(steps_n):
                    n_start, n_end = j * compute_tile_n, (j +
                                                          1) * compute_tile_n

                    rhs_q_slice = rhs_q_full[n_start:n_end, :]
                    rhs_scale_slice = rhs_scale_full[:, n_start:n_end]
                    if jnp.issubdtype(x_q_dtype, jnp.integer):
                        preferred_element_type = jnp.int32
                    else:
                        preferred_element_type = jnp.float32
                    dot_res = jax.lax.dot_general(
                        lhs_q,
                        rhs_q_slice,
                        (((1, ), (1, )), ((), ())),
                        preferred_element_type=preferred_element_type,
                    )
                    res = dot_res.astype(acc_dtype)
                    res = res * lhs_scale
                    res = res * rhs_scale_slice
                    if i == 0:
                        accumulators[j] = res
                    else:
                        accumulators[j] += res

            acc_block = jnp.concatenate(accumulators, axis=1)

            if not is_first_step:
                acc_block += acc_scratch[...]

            if is_last_step:
                out_ref[...] = acc_block.astype(out_ref.dtype)
            else:
                acc_scratch[...] = acc_block

        unfold_args((is_first_step, is_last_step), (), accum)

    kernel = pl.pallas_call(
        kernel,
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=0,
            in_specs=[
                pl.BlockSpec(
                    (batch_block_size, in_block_size),
                    lambda b, o, i: (b, i),
                    memory_space=pltpu.VMEM,
                ),  # x
                pl.BlockSpec(
                    (out_block_size, in_block_size),
                    lambda b, o, i: (o, i),
                    memory_space=pltpu.VMEM,
                ),  # w_q
                pl.BlockSpec(
                    (steps_k, 1, out_block_size),
                    lambda _, o, i: (i, 0, o),
                    memory_space=pltpu.VMEM,
                ),
            ],  # w_scale
            out_specs=pl.BlockSpec((batch_block_size, out_block_size),
                                   lambda b, o, i: (b, o)),
            scratch_shapes=[
                pltpu.VMEM((batch_block_size, out_block_size), jnp.bfloat16)
            ],
            grid=(n_batch, n_out, n_in),
        ),
        out_shape=jax.ShapeDtypeStruct((padded_n_batch, padded_n_out),
                                       x.dtype),
        compiler_params=pltpu.CompilerParams(
            dimension_semantics=("parallel", "parallel", "arbitrary"),
            vmem_limit_bytes=vmem_limit_bytes,
        ),
    )

    util.validate_inputs(
        x=x,
        w_q=w_q,
        w_scale=w_scale,
        x_abs_max=None,
        x_q_dtype=x_q_dtype,
        batch_block_size=batch_block_size,
        out_block_size=out_block_size,
        in_block_size=in_block_size,
    )

    # The named_scope is used for autotune.
    kernel_name = get_kernel_name(tuned_value)
    with jax.named_scope(kernel_name):
        out = kernel(x, w_q, w_scale)

    return out[:orig_n_batch, :orig_n_out]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from flax import nnx
from flax.typing import PRNGKey
from jax.sharding import Mesh

from tpu_inference.experimental.llama3_jax_stashed import (Llama3WeightLoader,
                                                           LlamaForCausalLM)


class MockParam:
    """A mock for a parameter used in the Llama model."""

    def __init__(self, shape=(32, 128)):
        self.value = SimpleNamespace(shape=shape)
        # The sharding spec is accessed during weight loading
        self.sharding = SimpleNamespace(spec=None)

    # Allow the mock parameter's value to be updated
    def __setattr__(self, name, value):
        if name == "value":
            self.__dict__[name] = value
        else:
            super().__setattr__(name, value)


class MockVllmConfig:
    """A mock VllmConfig sufficient for testing the Llama3 model."""

    def __init__(self,
                 model_name: str,
                 random_weights: bool = False,
                 tensor_parallelism: int = 1):
        self.model_config = SimpleNamespace(model=model_name,
                                            dtype="bfloat16",
                                            hf_overrides={},
                                            override_generation_config={})
        self.load_config = MagicMock()
        self.additional_config = {
            "random_weights": random_weights,
            "sharding": {
                "sharding_strategy": {
                    "tensor_parallelism": tensor_parallelism
                }
            }
        }

        # NOTE (jacobplatin): we could add a quantized KV cache test, but
        # we'll skip it for now.
        self.cache_config = MagicMock(cache_dtype="auto")


@pytest.fixture(scope="module")
def mesh():
    """
    Creates a mesh with all required axes for testing.
    FIX: The sharding logic expects 'data', 'model', and 'expert' axes.
    This creates a 3D mesh to satisfy the sharding rules, even on a single device.
    """
    if not jax.devices():
        pytest.skip("No JAX devices available for mesh creation.")

    devices = np.array(jax.local_devices())
    # Reshape devices into a 3D array to name 3 axes: data, model, and expert.
    # The 'model' and 'expert' axes will have a size of 1.
    num_devices = len(devices)
    device_mesh = devices.reshape((num_devices, 1, 1))

    with Mesh(device_mesh, axis_names=('data', 'model', 'expert')) as m:
        yield m


@pytest.fixture
def rng() -> PRNGKey:
    """Provides a reusable JAX PRNGKey."""
    return jax.random.PRNGKey(42)


@pytest.fixture
def mock_vllm_config_8b() -> MockVllmConfig:
    return MockVllmConfig(model_name="meta-llama/Llama-3-8B")


@pytest.fixture
def mock_vllm_config_70b() -> MockVllmConfig:
    return MockVllmConfig(model_name="meta-llama/Llama-3-70B-Instruct")


@pytest.fixture
def mock_vllm_config_unknown() -> MockVllmConfig:
    return MockVllmConfig(model_name="some-other-model")


# --- Test Cases ---


class TestLlamaForCausalLM:
    """Tests for the main LlamaForCausalLM model class."""

    def test_init_8b_variant(self, mock_vllm_config_8b, rng, mesh):
        """Tests correct parameter detection for the 8B model variant."""
        model = LlamaForCausalLM(mock_vllm_config_8b, rng, mesh)
        assert model.hidden_size == 4096
        assert "8b" in model.vllm_config.model_config.model.lower()

    def test_init_70b_variant(self, mock_vllm_config_70b, rng, mesh):
        """Tests correct parameter detection for the 70B model variant."""
        model = nnx.eval_shape(
            lambda: LlamaForCausalLM(mock_vllm_config_70b, rng, mesh))
        assert model.hidden_size == 8192
        assert "70b" in model.vllm_config.model_config.model.lower()

    def test_init_unknown_variant_raises_error(self, mock_vllm_config_unknown,
                                               rng, mesh):
        """Tests that an unknown model variant raises a ValueError."""
        with pytest.raises(ValueError,
                           match="Could not determine Llama3 variant"):
            LlamaForCausalLM(mock_vllm_config_unknown, rng, mesh)

    def test_create_model_with_random_weights(self, mock_vllm_config_8b, rng,
                                              mesh):
        """
        Tests that random weight initialization creates concrete, non-zero-variance arrays.
        """
        with jax.set_mesh(mesh):
            model = LlamaForCausalLM(vllm_config=mock_vllm_config_8b,
                                     rng=rng,
                                     mesh=mesh,
                                     force_random_weights=True)

            embedding_weight = model.embedder.input_embedding_table_VD.value
            attention_q_kernel = model.layers[0].attn.kernel_q_proj_DNH.value
            final_norm_scale = model.final_norm.scale.value

            assert isinstance(embedding_weight, jax.Array)
            assert isinstance(attention_q_kernel, jax.Array)
            assert isinstance(final_norm_scale, jax.Array)

            assert jnp.std(embedding_weight) > 0
            assert jnp.std(attention_q_kernel) > 0

            assert jnp.all(final_norm_scale == 1.0)

    @patch("tpu_inference.experimental.llama3_jax_stashed.Llama3WeightLoader")
    def test_load_weights_called_correctly(self, mock_loader_cls, rng, mesh):
        """Tests that the weight loader is called correctly for checkpoint loading."""
        vllm_config = MockVllmConfig(model_name="llama3-8b",
                                     random_weights=False)
        model = LlamaForCausalLM(vllm_config, rng, mesh)

        mock_loader_instance = MagicMock()
        mock_loader_cls.return_value = mock_loader_instance
        model.load_weights(rng, cache_dir="/tmp/cache")
        mock_loader_cls.assert_called_once_with(vllm_config=vllm_config,
                                                hidden_size=4096,
                                                attn_heads=32,
                                                num_key_value_heads=8,
                                                attn_head_dim=128)
        mock_loader_instance.load_weights.assert_called_once_with(model)


class TestLlama3WeightLoader:
    """Tests for the Llama3WeightLoader class."""

    @pytest.fixture
    def weight_loader(self):
        # Patch the superclass's setup to isolate the Llama3 loader's logic
        return Llama3WeightLoader(vllm_config=MockVllmConfig("test-model"),
                                  hidden_size=32,
                                  attn_heads=4,
                                  num_key_value_heads=2,
                                  attn_head_dim=8)

    def test_load_weights_transformation(self, weight_loader, rng, mesh):
        """Tests that weights are correctly reshaped, transposed, and loaded."""
        vllm_config = MockVllmConfig("llama3-8b-small-test",
                                     random_weights=False)

        # Create a model instance but override its config for the test.
        model = LlamaForCausalLM(vllm_config, rng, mesh)

        with patch(
                "tpu_inference.experimental.llama3_jax_stashed.load_hf_weights"
        ) as mock_load:
            # This will now pass after the code fix
            weight_loader.load_weights(model)

            # Assert that shard_put was called with the correctly transposed weight
            mock_load.assert_called_once()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from vllm.utils.network_utils import get_ip

from tpu_inference import envs
from tpu_inference.logger import init_logger

logger = init_logger(__name__)

# For multi-host usage only, to collect IP and port for all nodes.
_NODES_KV_IP_PORT = dict()


def set_node_kv_ip_port(ip_port: tuple[int, str, int]):
    global _NODES_KV_IP_PORT
    node_id, ip, port = ip_port
    _NODES_KV_IP_PORT[node_id] = (ip, port)


def get_kv_ips() -> str:
    if envs.TPU_MULTIHOST_BACKEND == "ray":
        num_nodes = len(_NODES_KV_IP_PORT)
        ips = []
        for node_id in range(num_nodes):
            ips.append(_NODES_KV_IP_PORT[node_id][0])
        return ips
    else:
        return get_host_ip()


def get_kv_ports() -> str:
    if envs.TPU_MULTIHOST_BACKEND == "ray":
        num_nodes = len(_NODES_KV_IP_PORT)
        ports = []
        for node_id in range(num_nodes):
            ports.append(_NODES_KV_IP_PORT[node_id][1])
        return ports
    else:
        return get_kv_transfer_port()


def get_host_ip() -> str:
    """Use `VLLM_HOST_IP` if set, otherwise use default network interface IP."""
    return get_ip()


def get_kv_transfer_port() -> str:
    port = os.getenv("TPU_KV_TRANSFER_PORT", "9100")
    return port


def get_side_channel_port() -> str:
    port = os.getenv("TPU_SIDE_CHANNEL_PORT", "9600")
    return port


def get_device_topology_order_id(local_devices, global_devices) -> int:
    """
    Calculates the topology order ID for the local device set within the global topology.

    This function determines the rank of the current host/process based on the
    coordinate of its TPU devices relative to all devices in the topology.

    Args:
        local_devices: A list of TpuDevice objects available to the current process.
        global_devices: A list of all TpuDevice objects in the global topology.

    Returns:
        The topology order ID (rank) of the local devices.
    """
    if not local_devices:
        raise ValueError("local_devices cannot be empty")
    if not global_devices:
        raise ValueError("global_devices cannot be empty")

    # 1. Find the 'anchor' (minimum coordinate) for the local devices.
    #    This represents the physical top-left corner of the local machine.
    local_anchor = min(d.coords for d in local_devices)

    # 2. Group global devices by process to find the anchor for EVERY process.
    process_anchors = {}
    for d in global_devices:
        pid = d.process_index
        # Update the minimum coordinate found for this process so far
        if pid not in process_anchors or d.coords < process_anchors[pid]:
            process_anchors[pid] = d.coords

    # 3. Sort the unique anchors to establish the canonical topology order.
    #    Tuples (x, y, z) sort lexicographically (x first, then y, then z).
    sorted_anchors = sorted(process_anchors.values())

    # 4. Return the index (rank) of the local anchor in the sorted list.
    try:
        return sorted_anchors.index(local_anchor)
    except ValueError:
        raise ValueError(
            f"Local devices: {local_devices} do not exist in the global device: {global_devices} list."
        )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from unittest.mock import MagicMock, patch

import pytest
from vllm.config import set_current_vllm_config
from vllm.distributed import cleanup_dist_env_and_memory
from vllm.distributed.parallel_state import (ensure_model_parallel_initialized,
                                             init_distributed_environment)
from vllm.engine.arg_utils import EngineArgs


@pytest.fixture(autouse=True)
def mock_get_pp_group():
    with patch("tpu_inference.distributed.jax_parallel_state.get_pp_group",
               return_value=MagicMock(is_first_rank=True,
                                      is_last_rank=True,
                                      rank_in_group=0,
                                      world_size=1)):
        yield


@pytest.fixture
def dist_init():
    engine_args = EngineArgs(
        model="Qwen/Qwen2-1.5B-Instruct",
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )

    vllm_config = engine_args.create_engine_config()

    with set_current_vllm_config(vllm_config):
        temp_file = tempfile.mkstemp()[1]
        init_distributed_environment(
            1,
            0,
            local_rank=0,
            distributed_init_method=f"file://{temp_file}",
            backend="gloo")
        ensure_model_parallel_initialized(1, 1)
        yield vllm_config
    cleanup_dist_env_and_memory(shutdown_ray=True)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import MagicMock, patch


# Mock VllmConfig and its nested configs to avoid dependencies on the actual
# classes, which can be complex to instantiate for testing.
class MockVllmConfig:

    def __init__(self):
        self.parallel_config = MagicMock()
        self.parallel_config.world_size = 4
        self.parallel_config.tensor_parallel_size = 2
        self.parallel_config.pipeline_parallel_size = 1
        self.parallel_config.ray_workers_use_nsight = False
        self.parallel_config.placement_group = None
        self.parallel_config.max_parallel_loading_workers = None

        self.sharding_config = MagicMock()
        self.sharding_config.total_devices = 2

        self.model_config = MagicMock()
        self.cache_config = MagicMock()
        self.lora_config = MagicMock()
        self.load_config = MagicMock()
        self.scheduler_config = MagicMock()
        self.speculative_config = MagicMock()
        self.prompt_adapter_config = MagicMock()
        self.observability_config = MagicMock()
        self.device_config = MagicMock()
        self.ec_transfer_config = MagicMock()


@patch(
    "vllm.v1.executor.ray_distributed_executor.RayDistributedExecutor.__init__",
    lambda x, y: None)
@patch("tpu_inference.executors.ray_distributed_executor.envs")
@patch("tpu_inference.executors.ray_distributed_executor.ray")
@patch("tpu_inference.executors.ray_distributed_executor.current_platform")
@patch("tpu_inference.executors.ray_distributed_executor.get_ip",
       return_value="127.0.0.1")
@patch("tpu_inference.executors.ray_distributed_executor.get_open_port",
       return_value=12345)
@patch(
    "tpu_inference.executors.ray_distributed_executor.available_resources_per_node"
)
@patch("tpu_inference.executors.ray_distributed_executor._wait_until_pg_ready")
class TestTpuRayDistributedExecutor(unittest.TestCase):

    def setUp(self):
        # Import the class under test inside the test method to ensure
        # patches are applied.
        from tpu_inference.executors.ray_distributed_executor import \
            RayDistributedExecutor
        self.RayDistributedExecutor = RayDistributedExecutor

        self.vllm_config = MockVllmConfig()
        # Reset placement group for each test as it might be modified.
        self.vllm_config.parallel_config.placement_group = None
        self.vllm_config.kv_transfer_config = None

    def test_init_executor_basic_flow(self, mock_wait_until_pg_ready,
                                      mock_avail_resources, mock_get_port,
                                      mock_get_ip, mock_platform, mock_ray,
                                      mock_envs):
        # --- Setup mocks ---
        mock_envs.VLLM_USE_RAY_COMPILED_DAG = True
        mock_envs.VLLM_USE_RAY_SPMD_WORKER = True
        mock_envs.VLLM_RAY_BUNDLE_INDICES = ""

        mock_platform.ray_device_key = "TPU"
        mock_platform.device_name = "tpu"
        mock_platform.device_control_env_var = "TPU_VISIBLE_CHIPS"
        mock_platform.additional_env_vars = []

        mock_ray.is_initialized.return_value = False
        mock_ray.nodes.return_value = [{"Resources": {"TPU": 4}}]
        mock_ray.get_runtime_context.return_value.get_node_id.return_value = "node_1"
        mock_avail_resources.return_value = {"node_1": {"TPU": 4}}

        mock_wait_until_pg_ready.return_value = None

        mock_placement_group = MagicMock()
        mock_placement_group.bundle_specs = [{"TPU": 1}] * 4
        mock_ray.util.placement_group.return_value = mock_placement_group

        mock_worker = MagicMock()
        mock_worker.get_node_and_gpu_ids.remote.return_value = [("node_1",
                                                                 [0, 1, 2, 3])]
        mock_ray.remote.return_value.remote.return_value = mock_worker

        # Simulate remote calls on the worker
        mock_ray.get.side_effect = [
            ["127.0.0.1"] * 4,  # worker_ips
            *[("node_1", [i]) for i in range(4)]  # worker_node_and_tpu_ids
        ]

        executor = self.RayDistributedExecutor(self.vllm_config)
        # Members of the parent class
        executor.uses_ray = True
        executor.vllm_config = self.vllm_config
        executor.parallel_config = self.vllm_config.parallel_config
        executor.collective_rpc = MagicMock()
        executor.collective_rpc.return_value = None

        # --- Initialization ---
        executor._init_executor()

        # --- Assertions ---
        mock_ray.init.assert_called_once()
        self.assertIsNotNone(executor.parallel_config.placement_group)
        self.assertEqual(len(executor.workers), 4)

    def test_initialize_ray_cluster_no_tpu_on_driver_raises_error(
            self, mock_wait_until_pg_ready, mock_avail_resources,
            mock_get_port, mock_get_ip, mock_platform, mock_ray, mock_envs):
        # --- Setup Mocks ---
        mock_platform.ray_device_key = "TPU"
        mock_platform.device_name = "tpu"

        mock_ray.is_initialized.return_value = False
        mock_ray.nodes.return_value = [{"Resources": {"TPU": 4}}]
        mock_ray.get_runtime_context.return_value.get_node_id.return_value = "driver_node"
        # Simulate no TPUs on the driver node
        mock_avail_resources.return_value = {
            "driver_node": {
                "CPU": 8
            },
            "worker_node": {
                "TPU": 4
            }
        }

        executor = self.RayDistributedExecutor(self.vllm_config)
        executor.vllm_config = self.vllm_config
        executor.parallel_config = self.vllm_config.parallel_config

        # --- Test and Assert ---
        with self.assertRaisesRegex(ValueError,
                                    "Current node has no TPU available"):
            executor._initialize_ray_cluster()

    def test_init_workers_ray_sorts_correctly(self, mock_wait_until_pg_ready,
                                              mock_avail_resources,
                                              mock_get_port, mock_get_ip,
                                              mock_platform, mock_ray,
                                              mock_envs):
        # --- Setup Mocks ---
        mock_envs.VLLM_RAY_BUNDLE_INDICES = ""
        mock_platform.ray_device_key = "TPU"
        mock_get_ip.return_value = "10.0.0.1"  # Driver IP

        mock_pg = MagicMock()
        mock_pg.bundle_specs = [{"TPU": 1}] * 4

        mock_workers = [MagicMock() for _ in range(4)]
        mock_ray.remote.return_value.return_value.remote.side_effect = mock_workers

        # Simulate IPs for workers created with ranks 0, 1, 2, 3
        worker_ips = ["10.0.0.2", "10.0.0.3", "10.0.0.1", "10.0.0.4"]
        mock_ray.get.side_effect = [
            worker_ips,  # worker_ips
            *[('node_1', ['0', '1', '2', '3']),
              ('node_2', ['4', '5', '6', '7']),
              ('node_3', ['8', '9', '10', '11']),
              ('node_4', ['12', '13', '14', '15'])]  # worker_node_and_tpu_ids
        ]

        executor = self.RayDistributedExecutor(self.vllm_config)
        executor.use_ray_spmd_worker = True
        executor.parallel_config = self.vllm_config.parallel_config
        executor.vllm_config = self.vllm_config
        executor.parallel_config.ray_workers_use_nsight = False
        executor.collective_rpc = MagicMock()
        executor.collective_rpc.return_value = None

        # --- Call method under test ---
        executor._init_workers_ray(mock_pg)

        # --- Assertions ---
        # Expected sorted order of workers: driver, then by IP
        # Original workers: 0 (10.0.0.2), 1 (10.0.0.3), 2 (10.0.0.1), 3 (10.0.0.2)
        # Sorted workers: 2 (driver), 0, 3 (same IP), 1
        self.assertEqual(executor.workers, [
            mock_workers[2], mock_workers[0], mock_workers[1], mock_workers[3]
        ])



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from jax.sharding import Mesh
from vllm.model_executor.layers.fused_moe import FusedMoE
from vllm.model_executor.layers.fused_moe.config import FusedMoEConfig

from tpu_inference import envs
from tpu_inference.kernels.fused_moe.v1.kernel import fused_ep_moe
from tpu_inference.layers.common.fused_moe_gmm import fused_moe_func
from tpu_inference.logger import init_logger

if TYPE_CHECKING:
    from tpu_inference.layers.vllm.process_weights.fused_moe_weights import \
        FusedMoEWeights

logger = init_logger(__name__)


class FusedMoEBackend(Enum):
    FUSED_MOE = "fused_moe"
    GMM_EP = "gmm_ep"
    GMM_TP = "gmm_tp"


def select_moe_backend(moe: FusedMoEConfig):
    if envs.USE_MOE_EP_KERNEL:
        if moe.use_ep:
            return FusedMoEBackend.FUSED_MOE
        logger.warning_once(
            "USE_MOE_EP_KERNEL=1 but expert parallelism is not "
            "enabled. Falling back to gmm implementation.")

    if moe.use_ep:
        return FusedMoEBackend.GMM_EP

    # Use default implementation.
    return FusedMoEBackend.GMM_TP


def fused_moe_apply(
    layer: FusedMoE,
    x: jax.Array,
    gating_output: jax.Array,
    weights: "FusedMoEWeights",
    moe_backend: FusedMoEBackend,
    mesh: Mesh,
    extra_backend_kwargs: dict,
) -> jax.Array:
    assert isinstance(layer, FusedMoE)
    if layer.scoring_func != "softmax":
        raise NotImplementedError("Only softmax is supported for scoring_func")

    with jax.named_scope(layer._get_name()):
        match moe_backend:
            case FusedMoEBackend.FUSED_MOE:
                subc_quant_w1_sz = None
                subc_quant_w2_sz = None
                if weights.w13_weight_scale is not None and weights.w2_weight_scale is not None:
                    padded_hidden_size = weights.w13_weight.shape[-2]
                    # NB: w13_weight_scale: (num_experts, 2, hidden_size // subc_quant_w1_sz, 1, intermediate_size)
                    assert padded_hidden_size % weights.w13_weight_scale.shape[
                        2] == 0
                    subc_quant_w1_sz = padded_hidden_size // weights.w13_weight_scale.shape[
                        2]
                    intermediate_size = weights.w13_weight.shape[-1]
                    # NB: w2_weight_scale: (num_experts, intermediate_size // subc_quant_w2_sz, 1, hidden_size)
                    assert intermediate_size % weights.w2_weight_scale.shape[
                        1] == 0
                    subc_quant_w2_sz = intermediate_size // weights.w2_weight_scale.shape[
                        1]

                actual_hidden_size = x.shape[-1]
                padding_size = weights.w13_weight.shape[-2] - actual_hidden_size
                x = jnp.pad(x, ((0, 0), (0, padding_size)))
                output = fused_ep_moe(
                    mesh=mesh,
                    tokens=x,
                    w1=weights.w13_weight,
                    w2=weights.w2_weight,
                    gating_output=gating_output,
                    top_k=layer.top_k,
                    renormalize_topk_logits=layer.renormalize,
                    act_fn=layer.activation,
                    subc_quant_w1_sz=subc_quant_w1_sz,
                    subc_quant_w2_sz=subc_quant_w2_sz,
                    w1_scale=weights.w13_weight_scale,
                    w2_scale=weights.w2_weight_scale,
                    b1=weights.w13_bias,
                    b2=weights.w2_bias,
                    **extra_backend_kwargs,
                )[:, :actual_hidden_size]
            case FusedMoEBackend.GMM_EP | FusedMoEBackend.GMM_TP:
                output = fused_moe_func(
                    hidden_states=x,
                    w1=weights.w13_weight,
                    w2=weights.w2_weight,
                    w1_scale=weights.w13_weight_scale,
                    w2_scale=weights.w2_weight_scale,
                    w1_bias=weights.w13_bias,
                    w2_bias=weights.w2_bias,
                    gating_output=gating_output,
                    topk=layer.top_k,
                    renormalize=layer.renormalize,
                    mesh=mesh,
                    use_ep=layer.use_ep,
                    activation=layer.activation,
                )

        return output
